#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020-2021 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide tools associated to anthropometric measurements and estimates.

Warning
-------
This module is in very early development and everything can still change.
Please don't use this module in production code.


Nomenclature
------------
Skin markers (either real or reconstructed):

- HeadVertex
- Sellion
- C7
- Suprasternale
- PubicSymphysis
- AnteriorSuperiorIliacSpineR/L
- PosteriorSuperiorIliacSpineR/L
- AcromionR/L
- LateralHumeralEpicondyleR/L
- MedialHumeralEpicondyleR/L
- UlnarStyloidR/L
- RadialStyloidR/L
- CarpalMetaHead2R/L
- CarpalMetaHead5R/L
- HipJointCenterR/L
- LateralFemoralEpicondyleR/L
- MedialFemoralEpicondyleR/L
- LateralMalleolusR/L
- MedialMalleolusR/L
- CalcaneusR/L
- NavicularR/L
- TarsalMetaHead1R/L
- TarsalMetaHead5R/L

Inferred joint centers:

- L5S1JointCenter
- C7T1JointCenter
- GlenohumeralJointCenterR/L
- ElbowJointCenterR/L
- WristJointCenterR/L
- KneeJointCenterR/L
- AnkleJointCenterR/L

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020-2021 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


from kineticstoolkit.decorators import unstable, directory
from kineticstoolkit import TimeSeries, geometry, config, kinematics, Player
from typing import Dict, Union, Sequence
import numpy as np
import pandas as pd
from warnings import warn


@unstable
def infer_joint_centers(
    markers: TimeSeries,
    /,
    segment: str,
    *,
    sex: str = "M",
) -> TimeSeries:
    """
    Infer joint centers based on anthropometric data.

    This function is aimed to be used on static acquisitions in quasi-neutral
    position.

    For the pelvis, creates L5S1JointCenter, HipJointCenterL and
    HipJointCenterR based on AnteriorSuperiorIliacSpineR/L,
    PosteriorSuperiorIliacSpineR/L and PubicSymphysis, following the method
    presented in Reed et al. (1999).

    For the thorax, creates C7T1JointCenter, GlenohumeralJointCenterL and
    GlenohumeralJointCenterR based on C7, L5S1JointCenter, Suprasternal,
    AcromionR and AcromionL, following the method presented in Reed et al.
    (1999). The scapulo-thoracic joint must be in a neutral
    position.

    For the extremities, creates:

    - ElbowJointCenterR/L as the midpoint between
      LateralHumeralEpicondyleR/L and MedialHumeralEpicondyleR;
    - WristJointCenterR as the midpoint between
      UlnarStyloidR/L and RadialStyloidR/L;
    - KneeJointCenterR/L as the midpoint between
      LateralFemoralEpicondyleR/L and MedialFemoralEpicondyleR/L;
    - AnkleJointCenterR/L as the midpoint between
      LateralMalleolusR/L and MedialMalleolusR/L.

    Parameters
    ----------
    markers
        TimeSeries that contain the trajectory of the required markers as
        Nx4 series.
    segment
        Can be either 'Pelvis', 'Thorax', or 'Extremities'. An error is raised
        in case of missing markers.
    sex
        Optional. Can be either 'M' or 'F'. The default is 'M'.

    Returns
    -------
    TimeSeries
        A TimeSeries that contains the inferred joint center trajectories.

    References
    ----------
    - Reed, M., Manary, M.A., Schneider, L.W., 1999. Methods for Measuring and
      Representing Automobile Occupant Posture. Presented at the International
      Congress &  Exposition, pp. 1999-01–0959.
      https://doi.org/10.4271/1999-01-0959

    """
    if segment == "Pelvis":
        return _infer_pelvis_joint_centers(markers, sex=sex)
    elif segment == "Thorax":
        return _infer_thorax_joint_centers(markers, sex=sex)
    elif segment == "Extremities":
        return _infer_extremity_joint_centers(markers)
    elif segment == "":
        out = markers.copy()
        for segment in ["Pelvis", "Thorax", "Extremities"]:
            try:
                out.merge(
                    infer_joint_centers(
                        out,
                        segment=segment,
                        sex=sex,
                    ),
                    in_place=True,
                )
            except ValueError:
                pass
        return out
    else:
        raise ValueError("Unrecognized segment name.")


def _infer_pelvis_joint_centers(
    markers: TimeSeries, sex: str = "M"
) -> TimeSeries:
    """Infer L5S1 and hip joint centers based on anthropometric data."""
    # Get the required markers
    try:
        rasis = markers.data["AnteriorSuperiorIliacSpineR"]
        lasis = markers.data["AnteriorSuperiorIliacSpineL"]
        rpsis = markers.data["PosteriorSuperiorIliacSpineR"]
        lpsis = markers.data["PosteriorSuperiorIliacSpineL"]
        sym = markers.data["PubicSymphysis"]
    except KeyError:
        raise ValueError(
            "Not enough markers to reconstruct L5S1 and hip joint centers."
        )

    # Create a local coordinate system at the anterior superior iliac spines
    # midpoint, according to Reed et al.
    masis = 0.5 * (rasis + lasis)
    lcs = geometry.create_frames(origin=masis, y=lasis - rasis, yz=masis - sym)

    # Calculate the markers in the local coordinate system
    local_rasis = geometry.get_local_coordinates(rasis, lcs)
    local_lasis = geometry.get_local_coordinates(lasis, lcs)
    local_rpsis = geometry.get_local_coordinates(rpsis, lcs)
    local_lpsis = geometry.get_local_coordinates(lpsis, lcs)
    local_sym = geometry.get_local_coordinates(sym, lcs)

    # Create a cluster using these locations
    cluster = {
        "AnteriorSuperiorIliacSpineR": np.nanmean(local_rasis, axis=0)[
            np.newaxis
        ],
        "AnteriorSuperiorIliacSpineL": np.nanmean(local_lasis, axis=0)[
            np.newaxis
        ],
        "PosteriorSuperiorIliacSpineR": np.nanmean(local_rpsis, axis=0)[
            np.newaxis
        ],
        "PosteriorSuperiorIliacSpineL": np.nanmean(local_lpsis, axis=0)[
            np.newaxis
        ],
        "PubicSymphysis": np.nanmean(local_sym, axis=0)[np.newaxis],
    }

    # Track the pelvis using this definition
    tracked_pelvis = kinematics.track_cluster(
        markers, cluster, include_lcs=True, lcs_name="Pelvis"
    )

    # Pelvis width
    pw = np.abs(np.nanmean(local_rasis[:, 1] - local_lasis[:, 1]))

    output = markers.copy(copy_data=False, copy_data_info=False)

    # Create L5S1JointCenter
    if sex == "F":
        local_position = np.array([[-0.289 * pw, 0.0, 0.172 * pw, 1.0]])
    elif sex == "M":
        local_position = np.array([[-0.264 * pw, 0.0, 0.126 * pw, 1.0]])
    else:
        raise ValueError("sex must be either 'M' or 'F'")

    output.data["L5S1JointCenter"] = geometry.get_global_coordinates(
        local_position, tracked_pelvis.data["Pelvis"]
    )

    # Create HipJointCenterR if not already included in markers
    if sex == "F":
        local_position = np.array(
            [[-0.197 * pw, -0.372 * pw, -0.270 * pw, 1.0]]
        )
    else:  # M
        local_position = np.array(
            [[-0.208 * pw, -0.361 * pw, -0.278 * pw, 1.0]]
        )
    output.data["HipJointCenterR"] = geometry.get_global_coordinates(
        local_position, tracked_pelvis.data["Pelvis"]
    )

    # Create HipJointCenterL if not already included in markers
    if sex == "F":
        local_position = np.array(
            [[-0.197 * pw, 0.372 * pw, -0.270 * pw, 1.0]]
        )
    else:  # M
        local_position = np.array(
            [[-0.208 * pw, 0.361 * pw, -0.278 * pw, 1.0]]
        )
    output.data["HipJointCenterL"] = geometry.get_global_coordinates(
        local_position, tracked_pelvis.data["Pelvis"]
    )

    return output


def _infer_thorax_joint_centers(
    markers: TimeSeries, sex: str = "M"
) -> TimeSeries:
    """Infer C7T1 and glenohumeral joint centers based on anthropom. data."""
    # Get the required markers
    try:
        c7 = markers.data["C7"]
        l5s1 = markers.data["L5S1JointCenter"]
        sup = markers.data["Suprasternale"]
        rac = markers.data["AcromionR"]
        lac = markers.data["AcromionL"]
    except KeyError:
        raise ValueError(
            "Not enough markers to reconstruct C7T1 and GH joint centers."
        )
        return

    if sex == "M":
        c7t1_angle = 8  # deg
        c7t1_ratio = 0.55
        gh_angle = -67  # deg
        gh_ratio = 0.42
    # elif sex == 'F':
    #     c7t1_angle = 14  # deg
    #     c7t1_ratio = 0.53
    #     gh_angle = -5  # deg
    #     gh_ratio = 0.53
    else:
        raise ValueError("sex must be 'M' for now.")

    # Create reference frames with x: X7-SUP, y: L5S1-C7, z: right
    c7sup = sup - c7
    c7_lcs = geometry.create_frames(origin=c7, x=(c7sup), xy=(c7 - l5s1))
    rac_lcs = geometry.create_frames(origin=rac, x=(c7sup), xy=(c7 - l5s1))
    lac_lcs = geometry.create_frames(origin=lac, x=(c7sup), xy=(c7 - l5s1))

    # Thorax width (tw)
    tw = np.sqrt(np.sum(np.nanmean(c7sup, axis=0) ** 2))

    # Local positions
    local_c7t1 = np.array(
        [
            [
                c7t1_ratio * tw * np.cos(np.deg2rad(c7t1_angle)),
                c7t1_ratio * tw * np.sin(np.deg2rad(c7t1_angle)),
                0.0,
                1.0,
            ]
        ]
    )

    local_rgh = np.array(
        [
            [
                gh_ratio * tw * np.cos(np.deg2rad(gh_angle)),
                gh_ratio * tw * np.sin(np.deg2rad(gh_angle)),
                0.0,
                1.0,
            ]
        ]
    )

    local_lgh = np.array(
        [
            [
                gh_ratio * tw * np.cos(np.deg2rad(gh_angle)),
                gh_ratio * tw * np.sin(np.deg2rad(gh_angle)),
                0.0,
                1.0,
            ]
        ]
    )

    # Return global positions
    output = markers.copy(copy_data=False, copy_data_info=False)
    output.data["C7T1JointCenter"] = geometry.get_global_coordinates(
        local_c7t1, c7_lcs
    )
    output.data["GlenohumeralJointCenterR"] = geometry.get_global_coordinates(
        local_rgh, rac_lcs
    )
    output.data["GlenohumeralJointCenterL"] = geometry.get_global_coordinates(
        local_lgh, lac_lcs
    )

    return output


def _infer_extremity_joint_centers(markers: TimeSeries) -> TimeSeries:
    """Infer extremity joint centers based on medial and lateral markers."""
    output = markers.copy(copy_data=False, copy_data_info=False)
    try:
        output.data["ElbowJointCenterR"] = 0.5 * (
            markers.data["LateralHumeralEpicondyleR"]
            + markers.data["MedialHumeralEpicondyleR"]
        )
    except KeyError:
        pass

    try:
        output.data["ElbowJointCenterL"] = 0.5 * (
            markers.data["LateralHumeralEpicondyleL"]
            + markers.data["MedialHumeralEpicondyleL"]
        )
    except KeyError:
        pass

    try:
        output.data["KneeJointCenterR"] = 0.5 * (
            markers.data["LateralFemoralEpicondyleR"]
            + markers.data["MedialFemoralEpicondyleR"]
        )
    except KeyError:
        pass

    try:
        output.data["KneeJointCenterL"] = 0.5 * (
            markers.data["LateralFemoralEpicondyleL"]
            + markers.data["MedialFemoralEpicondyleL"]
        )
    except KeyError:
        pass

    try:
        output.data["WristJointCenterR"] = 0.5 * (
            markers.data["RadialStyloidR"] + markers.data["UlnarStyloidR"]
        )
    except KeyError:
        pass

    try:
        output.data["WristJointCenterL"] = 0.5 * (
            markers.data["RadialStyloidL"] + markers.data["UlnarStyloidL"]
        )
    except KeyError:
        pass

    try:
        output.data["AnkleJointCenterR"] = 0.5 * (
            markers.data["LateralMalleolusR"]
            + markers.data["MedialMalleolusR"]
        )
    except KeyError:
        pass

    try:
        output.data["AnkleJointCenterL"] = 0.5 * (
            markers.data["LateralMalleolusL"]
            + markers.data["MedialMalleolusL"]
        )
    except KeyError:
        pass

    return output


@unstable
def track_local_coordinate_systems(
    markers: TimeSeries,
    /,
    segments: Union[str, Sequence[str]],
    *,
    method: int = 1,
) -> TimeSeries:
    """
    Create local coordinate system definitions based on static markers.

    Parameters
    ----------
    markers
        TimeSeries that contains marker trajectories as Nx4 during an action.

    segments
        Name of the segment. Can be 'Pelvis', 'Thorax', 'HeadNeck',
        'ArmR', 'ArmL', 'ForearmR', 'ForearmL', 'HandR', 'HandL', 'ThighR',
        'ThighL', 'LegR', 'LegL', 'FootR', or 'FootL'. An empty string or
        an empty list will try to process every segment.

    method
        Optional. Select one ot both methods to reconstruct the axial
        orientation of the arm as proposed by the ISB: 1 to use both humeral
        epicondyles, 2 to use the elbow center and the ulnar styloid. The
        default is 1. This parameter is ignored for segments other than
        ArmR and ArmL.

    Returns
    -------
    TimeSeries
        A TimeSeries that contains the segments' local position and orientation
        as series of Nx4x4 frames.

    """
    output = markers.copy(copy_data=False, copy_data_info=False)

    # If we have no segment or an empty list, go with every segment
    if len(segments) == 0:
        segments = [
            "Pelvis",
            "Thorax",
            "HeadNeck",
            "ArmR",
            "ArmL",
            "ForearmR",
            "ForearmL",
            "HandR",
            "HandL",
            "ThighR",
            "ThighL",
            "LegR",
            "LegL",
            "FootR",
            "FootL",
        ]

    # If we have a list of segments
    if not isinstance(segments, str):
        for segment in segments:
            output.merge(
                track_local_coordinate_systems(
                    markers, segment, method=method
                ),
                in_place=True,
            )
        return output

    # If we have a single segment
    segment = segments

    if segment == "Pelvis":
        try:
            rasis = markers.data["AnteriorSuperiorIliacSpineR"]
            lasis = markers.data["AnteriorSuperiorIliacSpineL"]
            rpsis = markers.data["PosteriorSuperiorIliacSpineR"]
            lpsis = markers.data["PosteriorSuperiorIliacSpineL"]
            l5s1 = markers.data["L5S1JointCenter"]
        except KeyError:
            raise ValueError(
                "Not enough markers to create the pelvis coordinate system."
            )

        mpsis = 0.5 * (rpsis + lpsis)
        output.data[segment] = geometry.create_frames(
            origin=l5s1,
            z=(rasis - lasis),
            xz=(0.5 * (rasis + lasis) - mpsis),
        )

    elif segment == "Thorax":
        try:
            c7t1 = markers.data["C7T1JointCenter"]
            sup = markers.data["Suprasternale"]
            l5s1 = markers.data["L5S1JointCenter"]
        except KeyError:
            raise ValueError(
                "Not enough markers to create the thorax coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=c7t1,
            y=(c7t1 - l5s1),
            xy=(sup - l5s1),
        )

    elif segment == "HeadNeck":
        try:
            c7t1 = markers.data["C7T1JointCenter"]
            sel = markers.data["Sellion"]
            hv = markers.data["HeadVertex"]
        except KeyError:
            raise ValueError(
                "Not enough markers to create the thorax coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=c7t1,
            y=(hv - c7t1),
            xy=(sel - c7t1),
        )

    elif segment in ["ArmR", "ArmL"]:
        side = segment[-1]
        try:
            elbow_center = markers.data[f"ElbowJointCenter{side}"]
            gh = markers.data[f"GlenohumeralJointCenter{side}"]
            if side == "R":
                r_ep = markers.data["LateralHumeralEpicondyleR"]
                l_ep = markers.data["MedialHumeralEpicondyleR"]
            elif side == "L":
                r_ep = markers.data["MedialHumeralEpicondyleL"]
                l_ep = markers.data["LateralHumeralEpicondyleL"]
            else:
                raise ValueError("side must be either 'R' or 'L'")

        except KeyError:
            raise ValueError(
                "Not enough markers to create the arm coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=gh,
            y=(gh - elbow_center),
            yz=(r_ep - l_ep),
        )

    elif segment in ["ForearmR", "ForearmL"]:
        side = segment[-1]
        try:
            elbow_center = markers.data[f"ElbowJointCenter{side}"]
            wrist_center = markers.data[f"WristJointCenter{side}"]
            if side == "R":
                r_st = markers.data["RadialStyloidR"]
                l_st = markers.data["UlnarStyloidR"]
            elif side == "L":
                r_st = markers.data["UlnarStyloidL"]
                l_st = markers.data["RadialStyloidL"]
            else:
                raise ValueError("side must be either 'R' or 'L'")
        except KeyError:
            raise ValueError(
                "Not enough markers to create the forearm coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=elbow_center,
            y=(elbow_center - wrist_center),
            yz=(r_st - l_st),
        )

    elif segment in ["HandR", "HandL"]:
        side = segment[-1]
        try:
            wrist_center = markers.data[f"WristJointCenter{side}"]
            if side == "R":
                r_meta = markers.data["CarpalMetaHead2R"]
                l_meta = markers.data["CarpalMetaHead5R"]
            elif side == "L":
                r_meta = markers.data["CarpalMetaHead5L"]
                l_meta = markers.data["CarpalMetaHead2L"]
            else:
                raise ValueError("side must be either 'R' or 'L'")
        except KeyError:
            raise ValueError(
                "Not enough markers to create the hand coordinate system."
            )

        meta_center = 0.5 * (r_meta + l_meta)
        output.data[segment] = geometry.create_frames(
            origin=wrist_center,
            y=(wrist_center - meta_center),
            yz=(r_meta - l_meta),
        )

    elif segment in ["ThighR", "ThighL"]:
        side = segment[-1]
        try:
            hip_center = markers.data[f"HipJointCenter{side}"]
            knee_center = markers.data[f"KneeJointCenter{side}"]
            if side == "R":
                r_ep = markers.data["LateralFemoralEpicondyleR"]
                l_ep = markers.data["MedialFemoralEpicondyleR"]
            elif side == "L":
                r_ep = markers.data["MedialFemoralEpicondyleL"]
                l_ep = markers.data["LateralFemoralEpicondyleL"]
            else:
                raise ValueError("side must be either 'R' or 'L'")
        except KeyError:
            raise ValueError(
                "Not enough markers to create the thigh coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=hip_center,
            y=(hip_center - knee_center),
            yz=(r_ep - l_ep),
        )

    elif segment in ["LegR", "LegL"]:
        side = segment[-1]
        try:
            knee_center = markers.data[f"KneeJointCenter{side}"]
            ankle_center = markers.data[f"AnkleJointCenter{side}"]
            if side == "R":
                r_mal = markers.data["LateralMalleolusR"]
                l_mal = markers.data["MedialMalleolusR"]
            elif side == "L":
                r_mal = markers.data["MedialMalleolusL"]
                l_mal = markers.data["LateralMalleolusL"]
            else:
                raise ValueError("side must be either 'R' or 'L'")
        except KeyError:
            raise ValueError(
                "Not enough markers to create the leg coordinate system."
            )

        output.data[segment] = geometry.create_frames(
            origin=knee_center,
            y=(knee_center - ankle_center),
            yz=(r_mal - l_mal),
        )

    elif segment in ["FootR", "FootL"]:
        side = segment[-1]
        try:
            ankle_center = markers.data[f"AnkleJointCenter{side}"]
            calc = markers.data[f"Calcaneus{side}"]
            if side == "R":
                r_meta = markers.data["TarsalMetaHead5R"]
                l_meta = markers.data["TarsalMetaHead1R"]
            elif side == "L":
                r_meta = markers.data["TarsalMetaHead1L"]
                l_meta = markers.data["TarsalMetaHead5L"]
            else:
                raise ValueError("side must be either 'R' or 'L'")
        except KeyError:
            raise ValueError(
                "Not enough markers to create the foot coordinate system."
            )

        # Create the local coordinate system
        meta_center = 0.5 * (r_meta + l_meta)
        output.data[segment] = geometry.create_frames(
            origin=ankle_center,
            x=(meta_center - calc),
            xz=(r_meta - l_meta),
        )

    return output


@unstable
def estimate_center_of_mass(
    markers: TimeSeries,
    /,
    segments: Union[str, Sequence[str]],
    *,
    sex: str = "M",
) -> TimeSeries:
    """
    Estimate the segments' center of mass based on anthropometric data.

    Parameters
    ----------
    markers
        TimeSeries that contains marker trajectories as Nx4 during an action.

    segments
        Name of the segment. Can be 'Pelvis', 'Thorax', 'HeadNeck',
        'ArmR', 'ArmL', 'ForearmR', 'ForearmL', 'HandR', 'HandL', 'ThighR',
        'ThighL', 'LegR', 'LegL', 'FootR', or 'FootL'. An empty string or
        an empty list will try to process every segment.

    sex
        Optional. Either 'M' or 'F'. The default is 'M'.

    Returns
    -------
    TimeSeries
        A TimeSeries with the trajectory of the segments' centers of mass,
        named {segment}CenterOfMass.

    """
    markers = markers.copy()  # We will add markers to it so copy it first.
    output = markers.copy(copy_data=False, copy_data_info=False)

    # If we have no segment or an empty list, go with every segment
    if len(segments) == 0:
        segments = [
            "Pelvis",
            "Thorax",
            "HeadNeck",
            "ArmR",
            "ArmL",
            "ForearmR",
            "ForearmL",
            "HandR",
            "HandL",
            "ThighR",
            "ThighL",
            "LegR",
            "LegL",
            "FootR",
            "FootL",
        ]

    # If we have a list of segments
    if not isinstance(segments, str):
        for segment in segments:
            output.merge(
                estimate_center_of_mass(markers, segment),
                in_place=True,
            )
        return output

    # From here we have a single segment
    segment = segments

    # Decompose segment name and side
    if segment not in ["Pelvis", "Thorax", "HeadNeck"]:
        side = segment[-1]
        segment = segment[0:-1]
    else:
        side = ""

    # Calculate the local coordinate system for this segment
    lcs = track_local_coordinate_systems(markers, segments=(segment + side))

    df = INERTIAL_VALUES["Dumas2007"]
    # Search the inertial value tables for the given segment and sex
    _ = df.loc[(df["Segment"] == segment) & (df["Gender"] == sex)]
    constants = _.to_dict("records")[0]

    # Add possible missing markers
    if "ProjectedHipJointCenter" in [
        constants["LengthPoint1"],
        constants["LengthPoint2"],
    ]:
        # Calculate ProjectedHipJointCenter
        local_rhip = geometry.get_local_coordinates(
            markers.data["HipJointCenterR"], lcs.data["Pelvis"]
        )
        local_rhip[:, 2] = 0  # Projection in sagittal plane
        local_lhip = geometry.get_local_coordinates(
            markers.data["HipJointCenterL"], lcs.data["Pelvis"]
        )
        local_lhip[:, 2] = 0  # Projection in sagittal plane
        local_hips = 0.5 * (local_rhip + local_lhip)

        markers.data[
            "ProjectedHipJointCenter"
        ] = geometry.get_global_coordinates(local_hips, lcs.data["Pelvis"])

    if "CarpalMetaHeadM25" in [
        constants["LengthPoint1"],
        constants["LengthPoint2"],
    ]:
        # Calculate midpoint of carpal meat head 2 and 5
        markers.data[f"CarpalMetaHeadM25{side}"] = 0.5 * (
            markers.data[f"CarpalMetaHead2{side}"]
            + markers.data[f"CarpalMetaHead5{side}"]
        )

    if "TarsalMetaHeadM15" in [
        constants["LengthPoint1"],
        constants["LengthPoint2"],
    ]:
        # Calculate midpoint of carpal meat head 2 and 5
        markers.data[f"TarsalMetaHeadM15{side}"] = 0.5 * (
            markers.data[f"TarsalMetaHead1{side}"]
            + markers.data[f"TarsalMetaHead5{side}"]
        )

    # Calculate the segment length
    segment_length = np.sqrt(
        np.sum(
            np.nanmean(
                markers.data[constants["LengthPoint2"] + side]
                - markers.data[constants["LengthPoint1"] + side],
                axis=0,
            )
            ** 2
        )
    )

    # Add COM output
    output.data[
        f"{segment}{side}CenterOfMass"
    ] = geometry.get_global_coordinates(
        np.array(
            [
                [
                    segment_length * constants["RelComX"],
                    segment_length * constants["RelComY"],
                    segment_length * constants["RelComZ"],
                    1.0,
                ]
            ]
        ),
        lcs.data[segment + side],
    )

    return output


@unstable
def estimate_global_center_of_mass(
    coms: TimeSeries, /, sex: str = "M"
) -> TimeSeries:
    """
    Estimate the global center of mass.

    Parameters
    ----------
    coms
        A TimeSeries with the trajectory of every segment's center of mass.
        The segments must be in this list: 'PelvisCenterOfMass',
        'ThoraxCenterOfMass', 'HeadNeckCenterOfMass',
        'ArmRCenterOfMass', 'ArmLCenterOfMass',
        'ForearmRCenterOfMass', 'ForearmLCenterOfMass',
        'HandRCenterOfMass', 'HandLCenterOfMass',
        'ThighRCenterOfMass', 'ThighLCenterOfMass',
        'LegRCenterOfMass', 'LegLCenterOfMass',
        'FootRCenterOfMass', 'FootLCenterOfMass'.
    sex
        Optional. Either 'M' or 'F'. The default is 'M'.

    Returns
    -------
    ktk.TimeSeries
        A TimeSeries with a single element named 'GlobalCenterOfMass'.

    """
    inertial_data = INERTIAL_VALUES["Dumas2007"]

    out = coms.copy(copy_data=False, copy_data_info=False)
    out.data["GlobalCenterOfMass"] = np.zeros([out.time.shape[0], 4])

    cumulative_mass = 0.0

    for data in coms.data:
        segment_name = data.replace("CenterOfMass", "")
        if segment_name not in ["Thorax", "HeadNeck", "Pelvis"]:
            # The segment name is terminated by L or R. Remove it.
            segment_name = segment_name[0:-1]

        segment_rel_mass = inertial_data.loc[
            (inertial_data["Segment"] == segment_name)
            & (inertial_data["Gender"] == sex),
            "RelMass",
        ]

        if len(segment_rel_mass) == 1:

            out.data["GlobalCenterOfMass"] += (
                float(segment_rel_mass) * coms.data[data]
            )

            cumulative_mass += float(segment_rel_mass)

    out.data["GlobalCenterOfMass"] /= cumulative_mass
    out.data["GlobalCenterOfMass"][:, 3] = 1

    return out


# %% Constants

#: A link model to help kinematics visualization
LINKS = {
    "Pelvis": {
        "Links": [
            ["AnteriorSuperiorIliacSpineR", "AnteriorSuperiorIliacSpineL"],
            ["AnteriorSuperiorIliacSpineL", "PosteriorSuperiorIliacSpineL"],
            ["PosteriorSuperiorIliacSpineL", "PosteriorSuperiorIliacSpineR"],
            ["PosteriorSuperiorIliacSpineR", "AnteriorSuperiorIliacSpineR"],
            ["AnteriorSuperiorIliacSpineR", "HipJointCenterR"],
            ["PosteriorSuperiorIliacSpineR", "HipJointCenterR"],
            ["AnteriorSuperiorIliacSpineL", "HipJointCenterL"],
            ["PosteriorSuperiorIliacSpineL", "HipJointCenterL"],
            ["AnteriorSuperiorIliacSpineR", "PubicSymphysis"],
            ["AnteriorSuperiorIliacSpineL", "PubicSymphysis"],
            ["HipJointCenterR", "PubicSymphysis"],
            ["HipJointCenterL", "PubicSymphysis"],
            ["HipJointCenterL", "HipJointCenterR"],
        ],
        "Color": [0.25, 0.5, 0.25],
    },
    "Trunk": {
        "Links": [
            ["L5S1JointCenter", "C7T1JointCenter"],
            ["C7", "GlenohumeralJointCenterR"],
            ["C7", "GlenohumeralJointCenterL"],
            ["Suprasternale", "GlenohumeralJointCenterR"],
            ["Suprasternale", "GlenohumeralJointCenterL"],
            ["C7", "AcromionR"],
            ["C7", "AcromionL"],
            ["Suprasternale", "AcromionR"],
            ["Suprasternale", "AcromionL"],
            ["GlenohumeralJointCenterR", "AcromionR"],
            ["GlenohumeralJointCenterL", "AcromionL"],
        ],
        "Color": [0.5, 0.5, 0],
    },
    "HeadNeck": {
        "Links": [
            ["Sellion", "C7T1JointCenter"],
            ["HeadVertex", "C7T1JointCenter"],
            ["Sellion", "HeadVertex"],
        ],
        "Color": [0.5, 0.5, 0.25],
    },
    "UpperArms": {
        "Links": [
            ["GlenohumeralJointCenterR", "ElbowJointCenterR"],
            ["GlenohumeralJointCenterR", "LateralHumeralEpicondyleR"],
            ["GlenohumeralJointCenterR", "MedialHumeralEpicondyleR"],
            ["LateralHumeralEpicondyleR", "MedialHumeralEpicondyleR"],
            ["GlenohumeralJointCenterL", "ElbowJointCenterL"],
            ["GlenohumeralJointCenterL", "LateralHumeralEpicondyleL"],
            ["GlenohumeralJointCenterL", "MedialHumeralEpicondyleL"],
            ["LateralHumeralEpicondyleL", "MedialHumeralEpicondyleL"],
        ],
        "Color": [0.5, 0.25, 0],
    },
    "Forearms": {
        "Links": [
            ["ElbowJointCenterR", "WristJointCenterR"],
            ["RadialStyloidR", "LateralHumeralEpicondyleR"],
            ["UlnarStyloidR", "MedialHumeralEpicondyleR"],
            ["RadialStyloidR", "UlnarStyloidR"],
            ["ElbowJointCenterL", "WristJointCenterL"],
            ["RadialStyloidL", "LateralHumeralEpicondyleL"],
            ["UlnarStyloidL", "MedialHumeralEpicondyleL"],
            ["RadialStyloidL", "UlnarStyloidL"],
        ],
        "Color": [0.5, 0, 0],
    },
    "Hands": {
        "Links": [
            ["RadialStyloidR", "CarpalMetaHead2R"],
            ["UlnarStyloidR", "CarpalMetaHead5R"],
            ["CarpalMetaHead2R", "CarpalMetaHead5R"],
            ["RadialStyloidL", "CarpalMetaHead2L"],
            ["UlnarStyloidL", "CarpalMetaHead5L"],
            ["CarpalMetaHead2L", "CarpalMetaHead5L"],
        ],
        "Color": [0.5, 0, 0.25],
    },
    "Tighs": {
        "Links": [
            ["HipJointCenterR", "KneeJointCenterR"],
            ["HipJointCenterR", "LateralFemoralEpicondyleR"],
            ["HipJointCenterR", "MedialFemoralEpicondyleR"],
            ["LateralFemoralEpicondyleR", "MedialFemoralEpicondyleR"],
            ["HipJointCenterL", "KneeJointCenterL"],
            ["HipJointCenterL", "LateralFemoralEpicondyleL"],
            ["HipJointCenterL", "MedialFemoralEpicondyleL"],
            ["LateralFemoralEpicondyleL", "MedialFemoralEpicondyleL"],
        ],
        "Color": [0, 0.5, 0.5],
    },
    "Legs": {
        "Links": [
            ["AnkleJointCenterR", "KneeJointCenterR"],
            ["LateralMalleolusR", "LateralFemoralEpicondyleR"],
            ["MedialMalleolusR", "MedialFemoralEpicondyleR"],
            ["LateralMalleolusR", "MedialMalleolusR"],
            ["AnkleJointCenterL", "KneeJointCenterL"],
            ["LateralMalleolusL", "LateralFemoralEpicondyleL"],
            ["MedialMalleolusL", "MedialFemoralEpicondyleL"],
            ["LateralMalleolusL", "MedialMalleolusL"],
        ],
        "Color": [0, 0.25, 0.5],
    },
    "Feets": {
        "Links": [
            ["CalcaneusR", "TarsalMetaHead1R"],
            ["CalcaneusR", "TarsalMetaHead5R"],
            ["MedialMalleolusR", "TarsalMetaHead1R"],
            ["LateralMalleolusR", "TarsalMetaHead5R"],
            ["CalcaneusR", "MedialMalleolusR"],
            ["CalcaneusR", "LateralMalleolusR"],
            ["TarsalMetaHead1R", "TarsalMetaHead5R"],
            ["CalcaneusL", "TarsalMetaHead1L"],
            ["CalcaneusL", "TarsalMetaHead5L"],
            ["MedialMalleolusL", "TarsalMetaHead1L"],
            ["LateralMalleolusL", "TarsalMetaHead5L"],
            ["CalcaneusL", "MedialMalleolusL"],
            ["CalcaneusL", "LateralMalleolusL"],
            ["TarsalMetaHead1L", "TarsalMetaHead5L"],
        ],
        "Color": [0.25, 0.0, 0.75],
    },
}


# Load Inertial Values
INERTIAL_VALUES = {}

# Dumas2007
_ = pd.read_csv(config.root_folder + "/data/anthropometrics_dumas_2007.csv")
_[["RelIXY", "RelIXZ", "RelIYZ"]] = _[["RelIXY", "RelIXZ", "RelIYZ"]].applymap(
    lambda s: complex(s)
)
INERTIAL_VALUES["Dumas2007"] = _


module_locals = locals()


def __dir__():
    return directory(module_locals)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deprecated module. Please use the kineticstoolkit_pushrimkinetics extension.
"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit.filters as filters
import kineticstoolkit.cycles as cycles
from kineticstoolkit import TimeSeries
from kineticstoolkit.decorators import unstable, dead, directory, deprecated

import numpy as np
from numpy import sin, cos, pi
import pandas as pd
import warnings
import struct  # to unpack binary data from SmartWheels' txt files
from typing import Union, Optional, List


@deprecated(
    "0.8",
    "2024",
    (
        "Please install the kineticstoolkit_pushtimkinetics extension "
        "and use the equivalent "
        "ktk.ext.pushrimkinetics.read_smartwheel() function."
    ),
)
def read_file(filename: str, /, file_format: str = "") -> TimeSeries:
    """
    Read a file containing pushrim kinetics data.

    Parameters
    ----------
    filename
        Name of the file to open
    file_format
        Format of the file. Can be either:

        - 'smartwheel' (SmartWheel CSV file)
        - 'smartwheeltxt' (SmartWheel SD-Card TXT file)
        - 'racingwheel' (will change)

    """
    if file_format == "":
        warnings.warn(
            "file_format will need to be explicitely specified in "
            "future versions. Now using 'smartwheel'.",
            FutureWarning,
        )
        file_format = "smartwheel"

    if file_format == "smartwheel":

        dataframe = pd.read_csv(filename, delimiter=None, header=None)
        if dataframe.shape[1] == 1:  # Retry with ; as separator
            dataframe = pd.read_csv(filename, delimiter=";", header=None)

        data = dataframe.to_numpy()
        index = data[:, 1]
        time = np.arange(0, len(index)) / 240
        channels = data[:, 6:12]
        forces = data[:, 18:21]
        moments = data[:, 21:24]
        angle_deg = data[:, 3]
        angle_rad = np.unwrap(np.deg2rad(angle_deg))

        ts = TimeSeries(time=time)

        ts.data["Index"] = index
        ts.data["Channels"] = channels
        ts.data["Forces"] = np.block([[forces, np.zeros((len(index), 1))]])
        ts.data["Moments"] = np.block([[moments, np.zeros((len(index), 1))]])
        ts.data["Angle"] = angle_rad

        ts = ts.add_data_info("Channels", "Unit", "raw")
        ts = ts.add_data_info("Forces", "Unit", "N")
        ts = ts.add_data_info("Moments", "Unit", "Nm")
        ts = ts.add_data_info("Angle", "Unit", "rad")

    elif file_format == "racingwheel":

        dataframe = pd.read_csv(filename, delimiter=",")
        data = dataframe.to_numpy()
        time = data[:, 0]
        channels = data[:, 1:7]
        battery = data[:, 8]

        ts = TimeSeries(time=time)

        ts.data["Channels"] = channels
        ts = ts.add_data_info("Channels", "Unit", "raw")

        ts.data["Battery"] = battery
        ts = ts.add_data_info("Battery", "Unit", "raw")

    elif file_format == "smartwheeltxt":

        data = {
            "ch1": [],
            "ch2": [],
            "ch3": [],
            "ch4": [],
            "ch5": [],
            "ch6": [],
            "angle_ticks": [],
        }

        length = 0
        with open(filename, "rb") as fid:

            while True:
                try:
                    _ = fid.read(2)
                    data["ch1"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch2"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch3"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch4"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch5"].append(struct.unpack("h", fid.read(2))[0])
                    data["ch6"].append(struct.unpack("h", fid.read(2))[0])
                    data["angle_ticks"].append(
                        struct.unpack("i", fid.read(4))[0]
                    )
                    _ = fid.read(8)
                    length += 1
                except Exception:
                    break

        ch1 = np.array(data["ch1"][1:length])  # Remove 1st sample to be
        ch2 = np.array(data["ch2"][1:length])  # consistent with CSV file
        ch3 = np.array(data["ch3"][1:length])
        ch4 = np.array(data["ch4"][1:length])
        ch5 = np.array(data["ch5"][1:length])
        ch6 = np.array(data["ch6"][1:length])
        angle_ticks = np.array(data["angle_ticks"][1:length])

        # Keep only 12 least significant bytes
        ch1 = np.mod(ch1, 2**12)
        ch2 = np.mod(ch2, 2**12)
        ch3 = np.mod(ch3, 2**12)
        ch4 = np.mod(ch4, 2**12)
        ch5 = np.mod(ch5, 2**12)
        ch6 = np.mod(ch6, 2**12)

        # Convert angle in radian
        angle = angle_ticks / 4096 * 2 * np.pi

        ts = TimeSeries(time=np.linspace(0, (length - 1) / 240, length - 1))
        ts.data["Channels"] = np.concatenate(
            [
                ch1[:, np.newaxis],
                ch2[:, np.newaxis],
                ch3[:, np.newaxis],
                ch4[:, np.newaxis],
                ch5[:, np.newaxis],
                ch6[:, np.newaxis],
            ],
            axis=1,
        )
        ts.data["Angle"] = angle

        ts = ts.add_data_info("Channels", "Unit", "raw")
        ts = ts.add_data_info("Angle", "Unit", "rad")

    else:
        raise ValueError("Unknown file format.")

    return ts


@unstable
def find_recovery_indices(Mz: np.ndarray, /) -> np.ndarray:
    """
    Find recovery indices based on a vector of propulsion moments.

    This function analyzes the Mz moments to find which data correspond to
    pushes and which data correspond to recoveries. The method is very
    conservative on what could be considered as a recovery, so that every
    index returned by this function is almost certain to correspond to a
    recovery. This function is used by `pushrimkinetics.remove_sinusoids`
    to identify the instants with no hand contact. It should not be used to
    isolate the push and recovery phases (use `ktk.cycles.detect_cycles()`
    instead).

    Parameters
    ----------
    Mz
        Array that contains the propulsion moments in Nm.

    Returns
    -------
    np.ndarray
        Array of bools where each True represents recovery.

    See Also
    --------
    ktk.cycles.detect_cycles

    """
    Mz = Mz.copy()

    threshold = 2.24  # (Nm): max tolerance for the remaining values.

    while np.nanmax(Mz) - np.nanmin(Mz) > threshold:

        # Remove 1% of data that are the farthest to the median:

        # Sort data
        index_to_remove = np.argsort(np.abs(Mz - np.nanmedian(Mz)))
        sorted_Mz = Mz[index_to_remove]
        index_to_remove = index_to_remove[~np.isnan(sorted_Mz)]

        # Remove the 1% upper.
        index_to_remove = index_to_remove[
            int(0.99 * len(index_to_remove)) - 1 :
        ]

        # Assign nan to these data
        Mz[index_to_remove] = np.nan

    index = ~np.isnan(Mz)

    return index


@deprecated(
    "0.8",
    "2024",
    (
        "Please install the kineticstoolkit_pushtimkinetics extension "
        "and use the equivalent ktk.ext.remove_offsets() function."
    ),
)
def remove_offsets(
    kinetics: TimeSeries, baseline_kinetics: Optional[TimeSeries] = None
) -> TimeSeries:
    """
    Remove dynamic offsets in forces and moments.

    Parameters
    ----------
    kinetics
        TimeSeries that contains at least Forces, Moments and Angle data.
    baseline_kinetics
        Optional. TimeSeries that contains at least Forces and Moments data.
        This TimeSeries contains a baseline trial, where the wheelchair must be
        pushed by an operator and where no external force must be applied on
        the pushrims. If no baseline is provided, the baseline is calculated
        based on a detection of recoveries in the supplied kinetics
        TimeSeries.

    Returns
    -------
    TimeSeries
        A copy of the input TimeSeries, where sinusoids are removed from
        Forces and Moments data.

    References
    ----------
    F. Chénier, R. Aissaoui, C. Gauthier, and D. H. Gagnon,
    "Wheelchair pushrim kinetics measurement: A method to cancel
    inaccuracies due to pushrim weight and wheel camber," Medical
    Engineering and Physics, vol. 40, pp. 75--86, 2017.

    """
    kinetics = kinetics.copy()

    if baseline_kinetics is None:
        # Create baseline kinetics.
        recovery_index = find_recovery_indices(kinetics.data["Moments"][:, 2])
        f_ofs = np.hstack(
            (
                kinetics.data["Forces"][recovery_index, 0:3],
                kinetics.data["Moments"][recovery_index, 0:3],
            )
        )
        theta_baseline = kinetics.data["Angle"][recovery_index]

    else:
        # Use baseline kinetics.
        f_ofs = np.hstack(
            (
                baseline_kinetics.data["Forces"][:, 0:3],
                baseline_kinetics.data["Moments"][:, 0:3],
            )
        )
        theta_baseline = baseline_kinetics.data["Angle"][:]

    # Do the regression
    theta_baseline = theta_baseline[:, np.newaxis]
    q = np.hstack(
        (
            np.sin(theta_baseline),
            np.cos(theta_baseline),
            np.ones((len(theta_baseline), 1)),
        )
    )
    A = np.linalg.lstsq(q, f_ofs, rcond=None)
    A = A[0]

    # Apply the regression to forces and moments
    theta = kinetics.data["Angle"]
    theta = theta[:, np.newaxis]

    f = np.hstack(
        (kinetics.data["Forces"][:, 0:3], kinetics.data["Moments"][:, 0:3])
    )

    q = np.hstack((np.sin(theta), np.cos(theta), np.ones((len(theta), 1))))

    f = f - q @ A

    # Make the output timeseries
    kinetics.data["Forces"][:, 0:3] = f[:, 0:3]  # type: ignore
    kinetics.data["Moments"][:, 0:3] = f[:, 3:6]  # type: ignore

    return kinetics


@deprecated(
    "0.8",
    "2024",
    (
        "Please install the kineticstoolkit_pushtimkinetics extension "
        "and use the equivalent apply_calibration() function."
    ),
)
def calculate_forces_and_moments(
    kinetics: TimeSeries,
    /,
    gains: Union[np.ndarray, str],
    offsets: np.ndarray = np.zeros((6)),
    *,
    transducer: str = "force_cell",
    reference_frame: str = "wheel",
) -> TimeSeries:
    """
    Calculate pushrim forces and moments based on raw channel values.

    For standard force cells (with each channel being a raw value
    corresponding to Fx, Fy, Fz, Mx, My, Mz, respectively), calculates
    the forces and moments using a sensitivity matrix (gains) and an
    offset vector (offsets).

    For SmartWheel, calculates the forces and moments using a gain
    vector (gains) and an offset vector (offsets).

        [Fx, Fy, Fz, Mx, My, Mz] = gains @ channels + offsets

    Parameters
    ----------
    kinetics
        Input TimeSeries that must contain a 'Channels' key in its data dict.
    gains
        6x6 gain matrix (force_cell) or gain vector of length 6 (smartwheel).
    offsets
        Optional. Offset vector of length 6.
    transducer
        Optional. 'force_cell' or 'smartwheel'.
    reference_frame
        Optional. 'wheel' or 'hub'. 'wheel' to report the forces and moments
        into the local wheel's reference frame; 'hub' to compensate for the
        wheel rotation and match the reference frame used by the SmartWheel:
        x anteroposterior, y in the wheel plane, upward for non-camberred
        wheels, and z perpendicular to the wheel plane, outward.

    Returns
    -------
    TimeSeries
        A copy of the input TimeSeries, with the added 'Forces'
        and 'Moments' data keys.

    """
    # Calculate the forces and moments and add to the output
    if transducer == "smartwheel":

        # Calculate the rotation angle to apply to the calculated kinetics
        if reference_frame == "wheel":
            theta = np.array([0.0])
        elif reference_frame == "hub":
            theta = kinetics.data["Angle"]
        else:
            raise ValueError("reference_frame must be 'wheel' or 'hub'")

        # Extract channels and angle
        ch = kinetics.data["Channels"] - 2048

        # Calculate the forces and moments
        Fx = (
            gains[0]
            * (
                ch[:, 0] * sin(theta)
                + ch[:, 2] * sin(theta + 2 * pi / 3)
                + ch[:, 4] * sin(theta + 4 * pi / 3)
            )
            + offsets[0]
        )

        Fy = (
            gains[1]
            * (
                ch[:, 0] * cos(theta)
                + ch[:, 2] * cos(theta + 2 * pi / 3)
                + ch[:, 4] * cos(theta + 4 * pi / 3)
            )
            + offsets[1]
        )

        Fz = gains[2] * (ch[:, 1] + ch[:, 3] + ch[:, 5]) + offsets[2]

        Mx = (
            gains[3]
            * (
                ch[:, 1] * sin(theta)
                + ch[:, 3] * sin(theta + 2 * pi / 3)
                + ch[:, 5] * sin(theta + 4 * pi / 3)
            )
            + offsets[3]
        )

        My = (
            gains[4]
            * (
                ch[:, 1] * cos(theta)
                + ch[:, 3] * cos(theta + 2 * pi / 3)
                + ch[:, 5] * cos(theta + 4 * pi / 3)
            )
            + offsets[4]
        )

        Mz = gains[5] * (ch[:, 0] + ch[:, 2] + ch[:, 4]) + offsets[5]
        forces_moments = np.block(
            [
                Fx[:, np.newaxis],
                Fy[:, np.newaxis],
                Fz[:, np.newaxis],
                Mx[:, np.newaxis],
                My[:, np.newaxis],
                Mz[:, np.newaxis],
            ]
        )

    elif transducer == "force_cell":

        # Calculate the rotation angle to apply to the calculated kinetics
        if reference_frame == "wheel":
            theta = np.array([0.0])
        elif reference_frame == "hub":
            raise NotImplementedError(
                "hub reference_frame not implemented yet"
                "for force_cell transducers."
            )
        else:
            raise ValueError("reference_frame must be 'wheel' or 'hub'")

        n_frames = kinetics.data["Channels"].shape[0]

        forces_moments = np.empty((n_frames, 6))
        for i_frame in range(n_frames):
            forces_moments[i_frame] = (
                gains @ kinetics.data["Channels"][i_frame] + offsets
            )

    # Format these data in the output timeseries
    kinetics = kinetics.copy()

    kinetics.data["Forces"] = np.concatenate(
        [forces_moments[:, 0:3], np.zeros((forces_moments.shape[0], 1))],
        axis=1,
    )
    kinetics = kinetics.add_data_info("Forces", "Unit", "N")

    kinetics.data["Moments"] = np.concatenate(
        [forces_moments[:, 3:6], np.zeros((forces_moments.shape[0], 1))],
        axis=1,
    )
    kinetics = kinetics.add_data_info("Moments", "Unit", "Nm")

    return kinetics


@deprecated(
    "0.8",
    "2024",
    (
        "Please install the kineticstoolkit_pushtimkinetics extension "
        "and use the equivalent "
        "ktk.ext.pushrimkinetics.calculate_velocity() function."
    ),
)
def calculate_velocity(tsin: TimeSeries, /) -> TimeSeries:
    """
    Calculate velocity based on wheel angle.

    The velocity is calculated by deriving the angle using a 2nd order
    Savitzky-Golay filter of length 21. This filter has been experimentally
    validated to maximize the signal-to-noise ratio for a SmartWheel recording
    at 240 Hz. This function may change signature in the future to include
    other filtering options. To manually get the velocity using a custom
    filter, please see the ``ktk.filters`` module.

    Parameters
    ----------
    tsin
        TimeSeries that contains at least the data key 'Angle'.

    Returns
    -------
    TimeSeries
                A copy of the TimeSeries with the added data key 'Velocity'.

    See Also
    --------
    ktk.filters.butter : Butterworth filter for TimeSeries
    ktk.filters.savgol : Savitsky-golay filter for TimeSeries
    ktk.filters.deriv : Derivative filter for TimeSeries

    """
    tsangle = TimeSeries()
    tsangle.time = tsin.time
    tsangle.data["Angle"] = tsin.data["Angle"]
    tsvelocity = filters.savgol(
        tsangle, window_length=21, poly_order=2, deriv=1
    )
    tsout = tsin.copy()
    tsout.data["Velocity"] = tsvelocity.data["Angle"]
    try:
        tsout = tsout.add_data_info(
            "Velocity", "Unit", tsout.data_info["Angle"]["Unit"] + "/s"
        )
    except KeyError:
        pass
    return tsout


@deprecated(
    "0.8",
    "2024",
    (
        "Please install the kineticstoolkit_pushtimkinetics extension "
        "and use the equivalent "
        "ktk.ext.pushrimkinetics.calculate_power() function."
    ),
)
def calculate_power(tsin: TimeSeries, /) -> TimeSeries:
    """
    Calculate power based on wheel velocity and moment.

    Parameters
    ----------
    tsin
        TimeSeries that contains at least the data keys 'Velocity' and
        'Moments'. The units must be consistent (e.g., rad/s and Nm)

    Returns
    -------
    TimeSeries
        A copy of the TimeSeries with the added data key 'Power'.

    """
    tsout = tsin.copy()
    tsout.data["Power"] = tsout.data["Velocity"] * tsout.data["Moments"][:, 2]
    tsout = tsout.add_data_info("Power", "Unit", "W")
    return tsout


# Deprecated constants:

_deprecated_CALIBRATION_MATRICES = {}
_deprecated_CALIBRATION_MATRICES["SmartWheel_93"] = {
    "gains": np.array([-0.1080, 0.1080, 0.0930, 0.0222, -0.0222, 0.0234999]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_94"] = {
    "gains": np.array([-0.1070, 0.1070, 0.0960, 0.0222, -0.0222, 0.0230]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_123"] = {
    "gains": np.array([-0.106, 0.106, 0.094, 0.022, -0.022, 0.0234999]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_124"] = {
    "gains": np.array([-0.106, 0.106, 0.0949999, 0.0215, -0.0215, 0.0225]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_125"] = {
    "gains": np.array([-0.104, 0.104, 0.0979999, 0.0215, -0.0215, 0.0225]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_126"] = {
    "gains": np.array([-0.1059999, 0.1059999, 0.086, 0.021, -0.021, 0.023]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_126_S18"] = {
    "gains": np.array([-0.1083, 0.1109, 0.0898, 0.0211, -0.0194, 0.0214]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_179_S18"] = {
    "gains": np.array([-0.1399, 0.1091, 0.0892, 0.0240, -0.0222, 0.0241]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_180_S18"] = {
    "gains": np.array([-0.1069, 0.1091, 0.0932, 0.0240, -0.0226, 0.0238]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}
_deprecated_CALIBRATION_MATRICES["SmartWheel_181_S18"] = {
    "gains": np.array([-0.1152, 0.1095, 0.0791, 0.0229, -0.0197, 0.0220]),
    "offsets": np.array([0.0, 10.0, 0.0, 0.0, 0.0, 0.0]),
    "transducer": "smartwheel",
}


def __getattr__(name):
    if name == "CALIBRATION_MATRICES":
        warnings.warn(
            "These sample calibration matrices will be removed in " "2024.",
            FutureWarning,
        )
        return globals()["_deprecated_CALIBRATION_MATRICES"]
    raise AttributeError(f"module {__name__} has no attribute {name}")


module_locals = locals()


def __dir__():  # pragma: no cover
    return directory(module_locals)


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide functions to calculate inverse dynamics.

Warning
-------
This module is currently experimental and its API and behaviour could be
modified in the future.

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit.filters
from kineticstoolkit.decorators import unstable, directory

import numpy as np
from typing import Dict, List
from kineticstoolkit import TimeSeries

import warnings

default_filter_fc = 10  # Hz
default_filter_order = 2


@unstable
def get_anthropometrics(
    segment_name: str, total_mass: float
) -> Dict[str, float]:
    """
    Get anthropometric values for a given segment name.

    For the moment, only this table is available:

    D. A. Winter, Biomechanics and Motor Control of Human Movement,
    4th ed. University of Waterloo, Waterloo, Ontario, Canada,
    John Wiley & Sons, 2009.

    Parameters
    ----------
    segment_name
        The name of the segment, either:

        - 'Hand' (wrist axis to knuckle II of middle finger)
        - 'Forearm' (elbow axis to ulnar styloid)
        - 'UpperArm' (glenohumeral axis to elbow axis)
        - 'ForearmHand' (elbow axis to ulnar styloid)
        - 'TotalArm' (glenohumeral joint to ulnar styloid)
        - 'Foot' (lateral malleolus to head metatarsal II)
        - 'Leg' (femoral condyles to medial malleolus)
        - 'Thigh' (greater trochanter to femoral condyles)
        - 'FootLeg' (fomeral condyles to medial malleolus)
        - 'TotalLeg' (greater trochanter to medial malleolus)
        - 'TrunkHeadNeck' (greater trochanter to glenohumeral joint)
        - 'HeadArmsTrunk' (greater trochanter to glenohumeral joint)

    total_mass
        The total mass of the person, in kg.

    Returns
    -------
    Dict[str, float]
        A dict with the following keys:

        - 'Mass' : Mass of the segment, in kg.
        - 'COMProximalRatio' : Distance between the segment's center of
          mass and the proximal joint, as a ratio of the distance between
          both joints.
        - 'COMDistalRatio' : Distance between the segment's center of mass
          and the distal joint, as a ratio of the distance between
          both joints.
        - 'GyrationCOMRatio': Radius of gyration around the segment's center
          of mass, as a ratio of the distance between both joints.
        - 'GyrationProximalRatio': Radius of gyration around the segment's
           proximal joint, as a ratio of the distance between both joints.
        - 'GyrationDistalRatio': Radius of gyration around the segment's
           distal joint, as a ratio of the distance between both joints.

    """
    table = dict()
    table["Hand"] = [0.006, 0.506, 0.494, 0.297, 0.587, 0.577]
    table["Forearm"] = [0.016, 0.430, 0.570, 0.303, 0.526, 0.647]
    table["UpperArm"] = [0.028, 0.436, 0.564, 0.322, 0.542, 0.645]
    table["ForearmHand"] = [0.022, 0.682, 0.318, 0.468, 0.827, 0.565]
    table["TotalArm"] = [0.050, 0.530, 0.470, 0.368, 0.645, 0.596]
    table["Foot"] = [0.0145, 0.50, 0.50, 0.475, 0.690, 0.690]
    table["Leg"] = [0.0465, 0.433, 0.567, 0.302, 0.528, 0.643]
    table["Thigh"] = [0.100, 0.433, 0.567, 0.323, 0.540, 0.653]
    table["FootLeg"] = [0.061, 0.606, 0.394, 0.416, 0.735, 0.572]
    table["TotalLeg"] = [0.161, 0.447, 0.553, 0.326, 0.560, 0.650]
    table["TrunkHeadNeck"] = [0.578, 0.66, 0.34, 0.503, 0.830, 0.607]
    table["HeadArmsTrunk"] = [0.678, 0.626, 0.374, 0.496, 0.798, 0.621]

    out = dict()
    try:
        out["Mass"] = table[segment_name][0] * total_mass
        out["COMProximalRatio"] = table[segment_name][1]
        out["COMDistalRatio"] = table[segment_name][2]
        out["GyrationCOMRatio"] = table[segment_name][3]
        out["GyrationProximalRatio"] = table[segment_name][4]
        out["GyrationDistalRatio"] = table[segment_name][5]
        return out
    except KeyError:
        raise ValueError(f'The segment "{segment_name}" is not available.')


@unstable
def calculate_com_position(
    ts: TimeSeries, inertial_constants: Dict[str, float]
) -> TimeSeries:
    """
    Calculate the position of the segment's center of mass.

    Adds the data key 'COMPosition' to the TimeSeries based on the
    'COMProximalRatio' key of the dict inertial_constants.

    Parameters
    ----------
    ts
        A TimeSeries with the following data keys:

        - ProximalJointPosition (Nx4)
        - DistalJointPosition (Nx4)

    inertial_constants
        A dict with at least a 'COMProximalRatio' key, which is the distance
        between the segment's center of mass and the proximal joint, as a
        ratio of the distance between both joints.

    Returns
    -------
    TimeSeries
        A copy of the input timeseries plus the 'COMPosition' data
        key.

    """
    ts = ts.copy()

    ts.data["COMPosition"] = (
        inertial_constants["COMProximalRatio"]
        * (ts.data["DistalJointPosition"] - ts.data["ProximalJointPosition"])
        + ts.data["ProximalJointPosition"]
    )
    return ts


@unstable
def calculate_com_acceleration(
    ts: TimeSeries, filter_func: str, **kwargs
) -> TimeSeries:
    """
    Calculate the acceleration of the segment's center of mass.

    Adds the data key 'COMAcceleration' to the TimeSeries based on the
    specified filter function.

    Parameters
    ----------
    ts
        A TimeSeries with the at least a 'COMPosition' data key.

    filter_func
        'savgol': calculate the acceleration using the 2nd order coefficient
        of a moving polynomial.
        'butter': no-lag 2nd order filter followed by a centered derivate.

    window_length
        Only for Savistky-Golay filters (savgol).
        The length of the filter window. window_length must be a positive
        odd integer less or equal than the length of the TimeSeries.

    poly_order
        Only for Savistky-Golay filters (savgol).
        Optional. The order of the polynomial used to fit the samples.
        polyorder must be less than window_length. The default is 2.

    fc
        Only for Butterworth filters (butter).
        Cut-off frequency in Hz.
    order
        Only for Butterworth filters (butter).
        Optional. Order of the filter. The default is 2.

    Returns
    -------
    TimeSeries
        A copy of the input timeseries plus the 'COMAcceleration' data
        key.

    """
    ts = ts.copy()

    if filter_func == "savgol":
        if "window_length" not in kwargs:
            raise ValueError(
                "window_length must be specified for Savitzky-Golay filters"
            )
        if "poly_order" not in kwargs:
            kwargs["poly_order"] = 2

        ts_com = ts.get_subset("COMPosition")
        ts_acc = kineticstoolkit.filters.savgol(
            ts_com,
            window_length=kwargs["window_length"],
            poly_order=kwargs["poly_order"],
            deriv=2,
        )

        return ts

    elif filter_func == "butter":
        if "fc" not in kwargs:
            raise ValueError("fc must be specified for Butterworth filters")
        if "order" not in kwargs:
            kwargs["order"] = 2

        ts_com = ts.get_subset("COMPosition")
        ts_acc = kineticstoolkit.filters.butter(
            ts_com, fc=kwargs["fc"], order=kwargs["order"]
        )

        ts_acc = kineticstoolkit.filters.deriv(ts_acc, n=2)
        ts_acc = ts_acc.rename_data("COMPosition", "COMAcceleration")

        ts = ts.merge(ts_acc, resample=True)

        return ts

    else:
        raise ValueError("Unknown filter type")


@unstable
def calculate_segment_angles(ts: TimeSeries) -> TimeSeries:
    """
    Calculate the segment's projection angles in the three axes.

    Parameters
    ----------
    ts
        A TimeSeries with the following data keys:

        - ProximalJointPosition (Nx4)
        - DistalJointPosition (Nx4)

    Returns
    -------
    TimeSeries
        A copy of the input timeseries plus the 'SegmentAngles' data
        key, represented as an Nx3 numpy array.

    """
    ts = ts.copy()

    proximal_to_distal_joint_distance = (
        ts.data["DistalJointPosition"] - ts.data["ProximalJointPosition"]
    )

    segment_angle_x = np.arctan2(
        proximal_to_distal_joint_distance[:, 2],
        proximal_to_distal_joint_distance[:, 1],
    )
    segment_angle_y = np.arctan2(
        proximal_to_distal_joint_distance[:, 2],
        proximal_to_distal_joint_distance[:, 0],
    )
    segment_angle_z = np.arctan2(
        proximal_to_distal_joint_distance[:, 1],
        proximal_to_distal_joint_distance[:, 0],
    )
    ts.data["SegmentAngles"] = np.concatenate(
        (
            segment_angle_x[:, np.newaxis],
            segment_angle_y[:, np.newaxis],
            segment_angle_z[:, np.newaxis],
        ),
        axis=1,
    )

    return ts


@unstable
def calculate_segment_rotation_rates(
    ts: TimeSeries, filter_func: str, **kwargs
) -> TimeSeries:
    """
    Calculate the segments' projected angular velocities and accelerations.

    Parameters
    ----------
    ts
        A TimeSeries with at least the 'SegmentAngles' key (Nx3).

    filter_func
        'savgol': calculate the velocity and acceleration using the 1st and
        2nd order coefficients of a moving polynomial.
        'butter': no-lag 2nd order filter followed by centered derivates.

    window_length
        Only for Savitzky-Golay filters (savgol).
        The length of the filter window. window_length must be a positive
        odd integer less or equal than the length of the TimeSeries.

    poly_order
        Only for Savitzky-Golay filters (savgol).
        Optional. The order of the polynomial used to fit the samples.
        polyorder must be less than window_length. The default is 2.

    fc
        Only for Butterworth filters (butter).
        Cut-off frequency in Hz.

    order
        Only for Butterworth filters (butter).
        Optional. Order of the filter. The default is 2.

    Returns
    -------
    TimeSeries
        A copy of the input timeseries plus the 'AngularVelocity' and
        'AngularAcceleration' data keys (each Nx3).

    """
    ts = ts.copy()

    if filter_func == "savgol":
        if "window_length" not in kwargs:
            raise ValueError(
                "window_length must be specified for Savitzky-Golay filters"
            )
        if "poly_order" not in kwargs:
            kwargs["poly_order"] = 2

        ts_angle = ts.get_subset("SegmentAngles")
        ts_angvel = kineticstoolkit.filters.savgol(
            ts_angle,
            window_length=kwargs["window_length"],
            poly_order=kwargs["poly_order"],
            deriv=1,
        )
        ts_angacc = kineticstoolkit.filters.savgol(
            ts_angle,
            window_length=kwargs["window_length"],
            poly_order=kwargs["poly_order"],
            deriv=2,
        )
        ts.data["AngularVelocity"] = ts_angvel.data["SegmentAngles"]
        ts.data["AngularAcceleration"] = ts_angacc.data["SegmentAngles"]

        return ts

    elif filter_func == "butter":
        if "fc" not in kwargs:
            raise ValueError("fc must be specified for Butterworth filters")
        if "order" not in kwargs:
            kwargs["order"] = 2

        ts_angle = ts.get_subset("SegmentAngles")
        ts_filt = kineticstoolkit.filters.butter(
            ts_angle, fc=kwargs["fc"], order=kwargs["order"]
        )

        ts_vel = kineticstoolkit.filters.deriv(ts_filt, n=1)
        ts_vel = ts_vel.rename_data("SegmentAngles", "AngularVelocity")

        ts_acc = kineticstoolkit.filters.deriv(ts_filt, n=2)
        ts_acc = ts_acc.rename_data("SegmentAngles", "AngularAcceleration")

        ts = ts.merge(ts_vel, resample=True)
        ts = ts.merge(ts_acc, resample=True)

        return ts

    else:
        raise ValueError("Unknown filter type")


@unstable
def calculate_proximal_wrench(
    ts: TimeSeries, inertial_constants: Dict[str, float]
) -> TimeSeries:
    """
    Calculate the proximal wrench based on a TimeSeries.

    This function is based on R. Dumas, R. Aissaoui, and J. A. De Guise,
    "A 3D generic inverse dynamic method using wrench notation and quaternion
    algebra,” Comput Meth Biomech Biomed Eng, vol. 7, no. 3, pp. 159–166, 2004.

    Parameters
    ----------
    ts
        A TimeSeries with the following data keys:

        - ProximalJointPosition (Nx4)
        - DistalJointPosition (Nx4)
        - ForceApplicationPosition (Nx4)
        - DistalForces (Nx4)
        - DistalMoments (Nx4)

        and these ones (although they can be reconstructed automatically by
        the function, in which case warnings are issued):

        - COMPosition (Nx4)
        - COMAcceleration (Nx4)
        - SegmentAngles (Nx3)

    inertial_constants
        A dict that contains the following keys:

        - 'Mass': Mass of the segment, in kg.
        - 'GyrationCOMRatio': Radius of gyration around the segment's
          center of mass, as a ratio of the distance between
          both joints.

        This dict may be generated using the get_anthropometrics function.

    Returns
    -------
    TimeSeries
        A copy of the input timeseries plus these extra data keys:

        - 'ProximalForces' (Nx4)
        - 'ProximalMoments' (Nx4)
    """
    ts = ts.copy()

    n_frames = ts.time.shape[0]

    proximal_to_distal_joint_distance = (
        ts.data["DistalJointPosition"] - ts.data["ProximalJointPosition"]
    )

    ts.data["RadiusOfGyration"] = np.zeros((n_frames, 3))
    ts.data["RadiusOfGyration"][:, 0] = inertial_constants[
        "GyrationCOMRatio"
    ] * np.sqrt(
        proximal_to_distal_joint_distance[:, 1] ** 2
        + proximal_to_distal_joint_distance[:, 2] ** 2
    )
    ts.data["RadiusOfGyration"][:, 1] = inertial_constants[
        "GyrationCOMRatio"
    ] * np.sqrt(
        proximal_to_distal_joint_distance[:, 0] ** 2
        + proximal_to_distal_joint_distance[:, 2] ** 2
    )
    ts.data["RadiusOfGyration"][:, 2] = inertial_constants[
        "GyrationCOMRatio"
    ] * np.sqrt(
        proximal_to_distal_joint_distance[:, 0] ** 2
        + proximal_to_distal_joint_distance[:, 1] ** 2
    )

    # Center of mass position and acceleration
    if "COMPosition" not in ts.data:
        ts = calculate_com_position(ts, inertial_constants)
        warnings.warn(
            "COMPosition not found in input TimeSeries. I calculated it."
        )

    if "COMAcceleration" not in ts.data:
        ts = calculate_com_acceleration(
            ts,
            filter_func="butter",
            fc=default_filter_fc,
            order=default_filter_order,
        )
        warnings.warn(
            f"COMAcceleration not found in input TimeSeries. I calculated it "
            f"using a low-pass ButterWorth filter of order "
            f"{default_filter_order} at {default_filter_fc} Hz."
        )

    # Rotation angle, velocity and acceleration

    if ("AngularVelocity" not in ts.data) or (
        "AngularAcceleration" not in ts.data
    ):

        # Start by calculating the angles
        if "SegmentAngles" not in ts.data:
            ts = calculate_segment_angles(ts)
            warnings.warn(
                "SegmentAngles not found in input TimeSeries. I calculated it."
            )

        ts = calculate_segment_rotation_rates(
            ts,
            filter_func="butter",
            fc=default_filter_fc,
            order=default_filter_order,
        )
        warnings.warn(
            f"COMAcceleration not found in input TimeSeries. I calculated it "
            f"using a low-pass ButterWorth filter of order "
            f"{default_filter_order} at {default_filter_fc} Hz."
        )

    # Forces line of the wrench equation (16)
    a_i = ts.data["COMAcceleration"][:, 0:3]
    g = np.array([0, -9.81, 0])
    F_i_minus_1 = ts.data["DistalForces"][:, 0:3]

    # Moments line of the wrench equation (16)
    c_i = (
        ts.data["COMPosition"][:, 0:3]
        - ts.data["ProximalJointPosition"][:, 0:3]
    )
    d_i = (
        ts.data["ForceApplicationPosition"] - ts.data["ProximalJointPosition"]
    )[:, 0:3]

    segment_mass = inertial_constants["Mass"]
    I_i_temp = segment_mass * ts.data["RadiusOfGyration"][:, 0:3] ** 2
    # Diagonalize I_i:
    I_i = np.zeros((n_frames, 3, 3))
    I_i[:, 0, 0] = I_i_temp[:, 0]
    I_i[:, 1, 1] = I_i_temp[:, 1]
    I_i[:, 2, 2] = I_i_temp[:, 2]

    alpha_i = ts.data["AngularAcceleration"]
    omega_i = ts.data["AngularVelocity"]

    M_i_minus_1 = ts.data["DistalMoments"][:, 0:3]

    # Calculation of the proximal wrench
    proximal_wrench = np.zeros((n_frames, 6, 1))
    for i_frame in range(n_frames):

        skew_symmetric_c_i = np.array(
            [
                [0, -c_i[i_frame, 2], c_i[i_frame, 1]],
                [c_i[i_frame, 2], 0, -c_i[i_frame, 0]],
                [-c_i[i_frame, 1], c_i[i_frame, 0], 0],
            ]
        )

        skew_symmetric_d_i = np.array(
            [
                [0, -d_i[i_frame, 2], d_i[i_frame, 1]],
                [d_i[i_frame, 2], 0, -d_i[i_frame, 0]],
                [-d_i[i_frame, 1], d_i[i_frame, 0], 0],
            ]
        )

        matrix_1 = np.block(
            [
                [segment_mass * np.eye(3), np.zeros((3, 3))],
                [segment_mass * skew_symmetric_c_i, I_i[i_frame]],
            ]
        )

        matrix_2 = np.block([a_i[i_frame] - g, alpha_i[i_frame]])
        matrix_2 = matrix_2[:, np.newaxis]  # Convert 1d to column vector

        matrix_3 = np.hstack(
            [
                np.zeros(3),
                np.cross(omega_i[i_frame], I_i[i_frame] @ omega_i[i_frame]),
            ]
        )
        matrix_3 = matrix_3[:, np.newaxis]  # Convert 1d to column vector

        matrix_4 = np.block(
            [[np.eye(3), np.zeros((3, 3))], [skew_symmetric_d_i, np.eye(3)]]
        )

        matrix_5 = np.block([F_i_minus_1[i_frame], M_i_minus_1[i_frame]])
        matrix_5 = matrix_5[:, np.newaxis]  # Convert 1d to column vector

        proximal_wrench[i_frame] = (
            matrix_1 @ matrix_2 + matrix_3 + matrix_4 @ matrix_5
        )

    # Initialize to a series of vectors of length 4
    ts.data["ProximalForces"] = np.zeros((n_frames, 4))
    ts.data["ProximalMoments"] = np.zeros((n_frames, 4))
    # Assign the 3 first components of the vectors
    ts.data["ProximalForces"][:, 0:3] = proximal_wrench[:, 0:3, 0]
    ts.data["ProximalMoments"][:, 0:3] = proximal_wrench[:, 3:6, 0]

    return ts


module_locals = locals()


def __dir__():
    return directory(module_locals)



# Copyright 2016 Clay Flannigan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Python implementation of m-dimensional Iterative Closest Point method.

Author: Clay Flannigan
https://github.com/ClayFlannigan/icp

Python implementation of m-dimensional Iterative Closest Point method. ICP
finds a best fit rigid body transformation between two point sets.
Correspondence between the points is not assumed. Included is an SVD-based
least-squared best-fit algorithm for corresponding point sets.
"""

import numpy as np
from sklearn.neighbors import NearestNeighbors


def best_fit_transform(A, B):
    """
    Calculates the least-squares best-fit transform that maps corresponding points A to B in m spatial dimensions
    Input:
      A: Nxm numpy array of corresponding points
      B: Nxm numpy array of corresponding points
    Returns:
      T: (m+1)x(m+1) homogeneous transformation matrix that maps A on to B
      R: mxm rotation matrix
      t: mx1 translation vector
    """

    assert A.shape == B.shape

    # get number of dimensions
    m = A.shape[1]

    # translate points to their centroids
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)
    AA = A - centroid_A
    BB = B - centroid_B

    # rotation matrix
    H = np.dot(AA.T, BB)
    U, S, Vt = np.linalg.svd(H)
    R = np.dot(Vt.T, U.T)

    # special reflection case
    if np.linalg.det(R) < 0:
        Vt[m - 1, :] *= -1
        R = np.dot(Vt.T, U.T)

    # translation
    t = centroid_B.T - np.dot(R, centroid_A.T)

    # homogeneous transformation
    T = np.identity(m + 1)
    T[:m, :m] = R
    T[:m, m] = t

    return T, R, t


def nearest_neighbor(src, dst):
    """
    Find the nearest (Euclidean) neighbor in dst for each point in src
    Input:
        src: Nxm array of points
        dst: Nxm array of points
    Output:
        distances: Euclidean distances of the nearest neighbor
        indices: dst indices of the nearest neighbor
    """

    assert src.shape == dst.shape

    neigh = NearestNeighbors(n_neighbors=1)
    neigh.fit(dst)
    distances, indices = neigh.kneighbors(src, return_distance=True)
    return distances.ravel(), indices.ravel()


def icp(A, B, init_pose=None, max_iterations=20, tolerance=0.001):
    """
    The Iterative Closest Point method: finds best-fit transform that maps points A on to points B
    Input:
        A: Nxm numpy array of source mD points
        B: Nxm numpy array of destination mD point
        init_pose: (m+1)x(m+1) homogeneous transformation
        max_iterations: exit algorithm after max_iterations
        tolerance: convergence criteria
    Output:
        T: final homogeneous transformation that maps A on to B
        distances: Euclidean distances (errors) of the nearest neighbor
        i: number of iterations to converge
    """

    assert A.shape == B.shape

    # get number of dimensions
    m = A.shape[1]

    # make points homogeneous, copy them to maintain the originals
    src = np.ones((m + 1, A.shape[0]))
    dst = np.ones((m + 1, B.shape[0]))
    src[:m, :] = np.copy(A.T)
    dst[:m, :] = np.copy(B.T)

    # apply the initial pose estimation
    if init_pose is not None:
        src = np.dot(init_pose, src)

    prev_error = 0

    for i in range(max_iterations):
        # find the nearest neighbors between the current source and destination points
        distances, indices = nearest_neighbor(src[:m, :].T, dst[:m, :].T)

        # compute the transformation between the current source and nearest destination points
        T, _, _ = best_fit_transform(src[:m, :].T, dst[:m, indices].T)

        # update the current source
        src = np.dot(T, src)

        # check error
        mean_error = np.mean(distances)
        if np.abs(prev_error - mean_error) < tolerance:
            break
        prev_error = mean_error

    # calculate final transformation
    T, _, _ = best_fit_transform(A, src[:m, :].T)

    return T, distances, i



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Identify cycles and time-normalize data.

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import numpy as np
from kineticstoolkit.timeseries import TimeSeries, TimeSeriesEvent
from kineticstoolkit.decorators import directory
import warnings
from typing import List, Dict, Tuple, Sequence, Optional


def detect_cycles(
    ts: TimeSeries,
    data_key: str,
    *,
    event_names: Sequence[str] = ["phase1", "phase2"],
    thresholds: Sequence[float] = [0.0, 1.0],
    directions: Sequence[str] = ["rising", "falling"],
    min_durations: Sequence[float] = [0.0, 0.0],
    max_durations: Sequence[float] = [np.Inf, np.Inf],
    min_peak_heights: Sequence[float] = [-np.Inf, -np.Inf],
    max_peak_heights: Sequence[float] = [np.Inf, np.Inf],
    **kwargs,
) -> TimeSeries:
    """
    Detect cycles in a TimeSeries based on a dual threshold approach.

    This function detects biphasic cycles and identifies the transitions as
    new events in the output TimeSeries. These new events are named:

    - `event_name1`:
      corresponds to the start of phase 1
    - `event_name2`:
      corresponds to the start of phase 2
    - '_':
      corresponds to the end of the cycle.

    Warning
    -------
    This function, which has been introduced in 0.4, is still experimental and
    may change signature or behaviour in the future.

    Parameters
    ----------
    ts
        TimeSeries to analyze.
    data_key
        Name of the data key to analyze in the TimeSeries.
    event_names
        Optional. Name of the events to add in the output TimeSeries. Default
        is ['phase1', 'phase2'].
    thresholds
        Optional. Value to cross to register phase changes. Default is
        [0., 1.].
    directions
        Optional. Direction to cross thresholds to register phase changes.
        Either ['rising', 'falling'] or ['falling', 'rising']. Default is
        ['rising', 'falling'].
    min_durations
        Optional. Minimal phase durations in seconds. Default is [0., 0.].
    max_durations
        Optional. Maximal phase durations in seconds. Default is
        [np.Inf, np.Inf]
    min_peak_heights
        Optional. Minimal peak values to be reached in both phases. Default is
        [-np.Inf, -np.Inf].
    max_peak_heights
        Optional. Maximal peak values to be reached in both phases. Default is
        [np.Inf, np.Inf].

    Returns
    -------
    TimeSeries
        A copy of ts with the events added.

    """
    # lowercase directions[0] once
    directions[0] = directions[0].lower()  # type: ignore
    if directions[0] != "rising" and directions[0] != "falling":
        raise ValueError("directions[0] must be 'rising' or 'falling'")

    # Find the pushes
    time = ts.time
    data = ts.data[data_key]

    events = []

    is_phase1 = True

    for i in range(time.shape[0]):

        if directions[0] == "rising":
            crossing1 = data[i] >= thresholds[0]
            crossing2 = data[i] <= thresholds[1]
        else:
            crossing1 = data[i] <= thresholds[0]
            crossing2 = data[i] >= thresholds[1]

        if is_phase1 and crossing1:

            is_phase1 = False
            events.append(TimeSeriesEvent(time[i], event_names[0]))

        elif (not is_phase1) and crossing2:

            is_phase1 = True
            events.append(TimeSeriesEvent(time[i], event_names[1]))

    # Ensure that we start with event_name1 and that it's not on time0
    while (events[0].name != event_names[0]) or (events[0].time == time[0]):
        events = events[1:]

    # Remove cycles where criteria are not reached.
    valid_events = []

    for i_event in range(0, len(events) - 1, 2):
        time1 = events[i_event].time
        time2 = events[i_event + 1].time
        try:
            time3 = events[i_event + 2].time
        except IndexError:
            time3 = np.Inf

        sub_ts1 = ts.get_ts_between_times(time1, time2, inclusive=True)
        sub_ts2 = ts.get_ts_between_times(time1, time3, inclusive=True)

        if directions[0] == "rising":
            the_peak1 = np.max(sub_ts1.data[data_key])
            the_peak2 = np.min(sub_ts2.data[data_key])
        else:
            the_peak1 = np.min(sub_ts1.data[data_key])
            the_peak2 = np.max(sub_ts2.data[data_key])

        if (
            time2 - time1 >= min_durations[0]
            and time2 - time1 <= max_durations[0]
            and time3 - time2 >= min_durations[1]
            and time3 - time2 <= max_durations[1]
            and the_peak1 >= min_peak_heights[0]
            and the_peak1 <= max_peak_heights[0]
            and the_peak2 >= min_peak_heights[1]
            and the_peak2 <= max_peak_heights[1]
        ):
            # Save it.
            valid_events.append(events[i_event])
            valid_events.append(events[i_event + 1])
            if not np.isinf(time3):
                valid_events.append(TimeSeriesEvent(time3, "_"))

    # Form the output timeseries
    tsout = ts.copy()
    for event in valid_events:
        tsout = tsout.add_event(event.time, event.name)
    tsout.sort_events()

    return tsout


def time_normalize(
    ts: TimeSeries,
    /,
    event_name1: str,
    event_name2: str,
    *,
    n_points: int = 100,
    span: Optional[Sequence[int]] = None,
) -> TimeSeries:
    """
    Time-normalize cycles in a TimeSeries.

    This method time-normalizes the TimeSeries at each cycle defined by
    event_name1 and event_name2 on n_points. The time-normalized cycles are
    put end to end. For example, for a TimeSeries that contains three
    cycles, a time normalization with 100 points will give a TimeSeries
    of length 300. The TimeSeries' events are also time-normalized, including
    event_name1 but with event_name2 renamed as '_'.

    Parameters
    ----------
    ts
        The TimeSeries to analyze.
    event_name1
        The event name that correspond to the begin of a cycle.
    event_name2
        The event name that correspond to the end of a cycle.
    n_points
        Optional. The number of points of the output TimeSeries.
    span
        Optional. Specifies which normalized points to include in the output
        TimeSeries. See note below.

    Returns
    -------
    TimeSeries
        A new TimeSeries where each cycle has been time-normalized.

    Notes
    -----
    The span argument is experimental and has been introduced in version 0.4.
    Use it to define which normalized points to include in the output
    TimeSeries. For example, to normalize in percents and to include only data
    from 10 to 90% of each cycle, assign 100 to n_points and [10, 90] to span.
    The resulting TimeSeries will then be expressed in percents and wrap each
    80 points. It is also possible to include pre-cycle or post-cycle data.
    For example, to normalize in percents and to include 20% pre-cycle and 15%
    post-cycle, assign 100 to n_points and [-20, 15] to span. The resulting
    TimeSeries will then wrap each 135 points with the cycles starting at 20,
    155, etc. and ending at 119, 254, etc. For each cycle, events outside the
    0-100% spans are ignored.

    """
    # Optional span
    if span is None:
        span = [0, n_points]

    # Find the final number of cycles
    if len(ts.events) < 2:
        raise (ValueError("No cycle can be defined from these event names."))

    i_cycle = 0

    # Initialize the destination TimeSeries
    dest_ts = ts.copy()
    dest_ts.events = []
    if n_points == 100:
        dest_ts.time_info["Unit"] = "%"
    else:
        dest_ts.time_info["Unit"] = f"1/{n_points}"

    dest_data = {}  # type: Dict[str, List[np.ndarray]]
    dest_data_shape = {}  # type: Dict[str, Tuple[int, ...]]

    # Go through all cycles
    while True:

        # Get the begin time for this cycle
        begin_time = ts.get_event_time(event_name1, i_cycle)

        # Get the end time for this cycle
        end_cycle = 0
        end_time = ts.get_event_time(event_name2, end_cycle)
        while end_time <= begin_time:
            end_cycle += 1
            end_time = ts.get_event_time(event_name2, end_cycle)

        # We are done. Quit the loop.
        if np.isnan(begin_time) or np.isnan(end_time):
            break

        # Get the extended begin and end times considering relative_span
        extended_begin_time = begin_time + span[0] / n_points * (
            end_time - begin_time
        )
        extended_end_time = begin_time + span[1] / n_points * (
            end_time - begin_time
        )

        # Extract this cycle
        subts = ts.get_ts_between_times(
            extended_begin_time, extended_end_time, inclusive=True
        )
        subts_backup = subts.copy()

        if subts.time.shape[0] == 0:
            raise ValueError("")

        # Resample this cycle on span + 1
        # (and get the first points after)
        try:
            subts = subts.resample(
                np.linspace(
                    extended_begin_time,
                    extended_end_time,
                    span[1] - span[0] + 1,
                )
            )
        except ValueError:
            subts = subts_backup  # In case the try messed with subts
            subts = subts.resample(
                np.linspace(
                    extended_begin_time,
                    extended_end_time,
                    span[1] - span[0] + 1,
                ),
                fill_value="extrapolate",
            )
            warnings.warn(f"Cycle {i_cycle} has been extrapolated.")

        # Keep only the first points (the last one belongs to the next cycle)
        subts = subts.get_ts_between_indexes(
            0, span[1] - span[0] - 1, inclusive=True
        )

        # Keep only the events in the unextended span
        events = []
        for event in subts.events:
            if event.time >= begin_time and event.time < end_time:
                events.append(event)
        subts.events = events
        subts = subts.sort_events()

        # Separate start/end events from the other
        start_end_events = []
        other_events = []
        for event in subts.events:
            if event.name == event_name1 or event.name == event_name2:
                start_end_events.append(event)
            else:
                other_events.append(event)

        # Add event_name1 at the beginning and end (duplicates will be
        # cancelled at the end)
        dest_ts = dest_ts.add_event(
            -span[0] + i_cycle * (span[1] - span[0]), event_name1
        )
        dest_ts = dest_ts.add_event(
            -span[0] + n_points + i_cycle * (span[1] - span[0]), "_"
        )

        # Add the other events
        def time_to_normalized_time(time):
            """Resample the events times."""
            return (time - extended_begin_time) / (
                extended_end_time - extended_begin_time
            ) * (span[1] - span[0]) + i_cycle * (span[1] - span[0])

        for i_event, event in enumerate(other_events):

            # Resample
            new_time = time_to_normalized_time(event.time)
            dest_ts = dest_ts.add_event(new_time, event.name)

        # Add this cycle to dest_time and dest_data
        for key in subts.data:
            if key not in dest_data:
                dest_data[key] = []
                dest_data_shape[key] = ts.data[key].shape
            dest_data[key].append(subts.data[key])

        i_cycle += 1

    n_cycles = i_cycle

    # Put back dest_time and dest_data in dest_ts
    dest_ts.time = 1.0 * np.arange(n_cycles * (span[1] - span[0]))
    for key in ts.data:
        # Stack the data into a [cycle, percent, values] shape
        temp = np.array(dest_data[key])
        # Reshape to put all cycles end to end
        new_shape = list(dest_data_shape[key])
        new_shape[0] = n_cycles * (span[1] - span[0])
        dest_ts.data[key] = np.reshape(temp, new_shape)

    dest_ts = dest_ts.sort_events()
    return dest_ts


def stack(ts: TimeSeries, /, n_points: int = 100) -> Dict[str, np.ndarray]:
    """
    Stack time-normalized TimeSeries' data into a dict of arrays.

    This methods returns the data of a time-normalized TimeSeries as a dict
    where each key corresponds to a TimeSeries' data, and contains a numpy
    array where the first dimension is the cycle, the second dimension is the
    percentage of cycle, and the other dimensions are the data itself.

    Parameters
    ----------
    ts
        The time-normalized TimeSeries.
    n_points
        Optional. The number of points the TimeSeries has been time-normalized
        on.

    Returns
    -------
    Dict[str, np.ndarray]

    See Also
    --------
    kineticstoolkit.cycles.unstack

    """
    if np.mod(len(ts.time), n_points) != 0:
        raise (
            ValueError("It seems that this TimeSeries is not time-normalized.")
        )

    data = dict()
    for key in ts.data.keys():
        current_shape = ts.data[key].shape
        new_shape = [-1, n_points]
        for i in range(1, len(current_shape)):
            new_shape.append(ts.data[key].shape[i])
        data[key] = ts.data[key].reshape(new_shape, order="C")
    return data


def unstack(data: Dict[str, np.ndarray], /) -> TimeSeries:
    """
    Unstack time-normalized data from a dict of arrays to a TimeSeries.

    This methods creates a time-normalized TimeSeries by putting each cycle
    from the provided data dictionary end to end.

    Parameters
    ----------
    data
        A dict where each key contains a numpy array where the first dimension
        is the cycle, the second dimension is the percentage of cycle, and
        the other dimensions are the data itself.

    Returns
    -------
    TimeSeries

    See Also
    --------
    ktk.cycles.stack

    """
    ts = TimeSeries()
    for key in data.keys():
        current_shape = data[key].shape
        n_cycles = current_shape[0]
        n_points = current_shape[1]
        ts.data[key] = data[key].reshape([n_cycles * n_points], order="C")
    ts.time = np.arange(n_cycles * n_points)
    ts.time_info["Unit"] = ""
    return ts


# The stack_events function is working but commented for now, since I could not
# figure an obvious, undiscutable way to represent its output (use lists,
# TimeSeriesEvents, numpy arrays?). It's also unclear for me how to integrate
# with the standard stack function and its unstack counterpart.
#
# def stack_events(
#         ts: TimeSeries, /,
#         n_points: int = 100) -> Dict[str, np.ndarray]:
#     """
#     Stack time-normalized TimeSeries' events into a dict of arrays.

#     This methods returns the a dictionary where each key corresponds to an
#     event name, and contains a 2d numpy array that contains the event's
#     normalized time, with the first dimension being the cycle and the second
#     dimension being the occurrence of this event during this cycle.

#     Warning
#     -------
#     This function is currently experimental and may change signature and
#     behaviour in the future.

#     Parameters
#     ----------
#     ts
#         The time-normalized TimeSeries.
#     n_points
#         Optional. The number of points the TimeSeries has been
#         time-normalized on.

#     Returns
#     -------
#     Dict[str, np.ndarray]

#     Example
#     -------
#     >>> import kineticstoolkit.lab as ktk
#     >>> # Create a TimeSeries with different time-normalized events
#     >>> ts = ktk.TimeSeries(time=np.arange(400))  # 4 cycles of 100%
#     >>> ts = ts.add_event(9, 'event1')    # event1 at 9% of cycle 0
#     >>> ts = ts.add_event(110, 'event1')  # event1 at 10% of cycle 1
#     >>> ts = ts.add_event(312, 'event1')  # event1 at 12% of cycle 3
#     >>> ts = ts.add_event(382, 'event1')  # 2nd occ. event1 at 82% of cycle 3

#     >>> # Stack these events
#     >>> events = ktk.cycles.stack_events(ts)
#     >>> events['event1']
#     [[9.0], [10.0], [], [12.0, 82.0]]

#     """
#     ts = ts.copy()
#     ts.sort_events()

#     n_cycles = int(ts.time.shape[0] / n_points)
#     out = {}  # type: Dict[str, np.ndarray]

#     # Init
#     for event in ts.events:
#         if event.name not in out:
#             out[event.name] = [[] for i in range(n_cycles)]

#     for event in ts.events:
#         event_cycle = int(event.time / n_points)

#         out[event.name][event_cycle].append(np.mod(event.time, n_points))

#     return out


def most_repeatable_cycles(data: np.ndarray, /) -> List[int]:
    """
    Get the indexes of the most repeatable cycles in TimeSeries or array.

    This function returns an ordered list of the most repeatable to the least
    repeatable cycles.

    It works by recursively discarding the cycle than maximizes the
    root-mean-square error between the cycle and the average of every
    remaining cycle, until there are only two cycles remaining. The function
    returns a list that is the reverse order of cycle removal: first the two
    last cycles, then the last-removed cycle, and so on. If two cycles are as
    equivalently repeatable, they are returned in order of appearance.

    Note
    ----
    Cycles that include at least one NaN are excluded.


    Parameters
    ----------
    data
        Stacked time-normalized data to analyze, in the shape
        (n_cycles, n_points).

    Returns
    -------
    List[int]
        List of indexes corresponding to the cycles in most to least
        repeatable order.

    Example
    -------
    >>> import kineticstoolkit.lab as ktk
    >>> import numpy as np
    >>> # Create a data sample with four different cycles, the most different
    >>> # begin cycle 2 (cos instead of sin), then cycle 0.
    >>> x = np.arange(0, 10, 0.1)
    >>> data = np.array([[np.sin(x)], \
        [np.sin(x) + 0.14], \
        [np.cos(x) + 0.14], \
        [np.sin(x) + 0.15]])

    >>> ktk.cycles.most_repeatable_cycles(data)
    [1, 3, 0, 2]

    """
    data = data.copy()
    n_cycles = data.shape[0]
    out_cycles = []  # type: List[int]
    done_cycles = []  # type: List[int]  # Like out_cycles but includes NaNs

    # Exclude cycles with nans: put nans for all data of this cycle
    for i_cycle in range(n_cycles - 1, -1, -1):
        if np.isnan(np.sum(data[i_cycle])):
            data[i_cycle] = np.nan
            done_cycles.append(i_cycle)

    # Iteratively remove the cycle that is the most different from the
    # mean of the remaining cycles.
    while len(done_cycles) < n_cycles - 2:

        current_mean_cycle = np.nanmean(data, axis=0)

        rms = np.zeros(n_cycles)

        for i_curve in range(n_cycles - 1, -1, -1):
            rms[i_curve] = np.sqrt(
                np.mean(np.sum((data[i_curve] - current_mean_cycle) ** 2))
            )

        i_cycle = int(np.nanargmax(rms))
        out_cycles.append(i_cycle)
        done_cycles.append(i_cycle)
        data[i_cycle] = np.nan

    # Find the two remaining cycles
    set_all = set(range(n_cycles))
    set_done = set(done_cycles)
    remain = sorted(list(set_all - set_done))
    if len(remain) > 1:
        out_cycles.append(remain[1])
    if len(remain) > 0:
        out_cycles.append(remain[0])

    return out_cycles[-1::-1]


module_locals = locals()


def __dir__():
    return directory(module_locals)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide documentation tools to learn using Kinetics Toolkit.
"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2022 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import kineticstoolkit.config as config
import os
from kineticstoolkit.decorators import unstable


def download(filename: str, **kwargs) -> str:
    """
    Download example data and return its local file name.

    This function download example data from github. These data are the same
    that are used to generate the
    [documentation website](https://kineticstoolkit.uqam.ca).

    These example data are volatile; they are supplied only for the user to
    reproduce the tutorials. Therefore, these data may change according to
    changes in the tutorials.

    Parameters
    ----------
    filename
        A string that indicates which file to download.

    Returns
    -------
    str
        The file name (with complete path) of the downloaded sample data.

    """
    # Additional information for developers:
    # kwargs may include force_download=True, to force download from github.
    # In standard case, the local git version is used to save on download time.
    if "force_download" not in kwargs:
        kwargs["force_download"] = False

    try:
        import requests
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "The requests module is an optional dependency of Kinetics "
            "Toolkit. It must be installed to allow downloading data."
        )

    # Try to get the file locally from the local git repository
    path = config.root_folder + "/data/" + filename
    if os.path.exists(path) is True and kwargs["force_download"] is False:
        return path

    else:
        # Otherwise download it.
        file = requests.get(
            "https://github.com/felixchenier/kineticstoolkit/raw/master/data/"
            + filename
        )
        if file.status_code == 404:
            raise FileNotFoundError(
                f"'{filename}' could not be found on the server."
            )

        path = config.temp_folder + "/" + filename
        open(path, "wb").write(file.content)

        return path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides the Player class to visualize markers and rigid bodies in 3d.

The Player class is accessible directly from the toplevel Kinetics Toolkit
namespace (i.e., ktk.Player).
"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


from kineticstoolkit.timeseries import TimeSeries
from kineticstoolkit.decorators import directory

import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import animation
import numpy as np
from numpy import sin, cos
import time
import copy
from typing import Union, Sequence, Dict, Any, Optional

# To fit the new viewpoint on selecting a new marker
import scipy.optimize as optim


class Player:
    """
    A class that allows visualizing markers and rigid bodies in 3D.

    player = ktk.Player(parameters) creates and launches a Player
    instance.

    Parameters
    ----------
    *ts
        Contains the markers and rigid bodies to visualize, where each data
        key is either a marker position expressed as Nx4 array (N=time),
        or a rigid body pose expressed as a Nx4x4 array. Multiple TimeSeries
        can be provided, e.g., ktk.Player(markers, rigid_bodies)

    interconnections
        Optional. Each key corresponds to an inerconnection between markers,
        where one interconnection is another dict with the following keys:

        - Links: list of list of 2 str, where each str is a marker
          name. For example, to link Marker1 to Marker2 and
          Marker1 to Marker3, Links would be:
          [['Marker1', 'Marker2'], ['Marker1', 'Marker3']]

        - Color: character or tuple that represents the color of the
          link. Color must be a valid value for matplotlib's
          plots.

    current_index
        Optional. Sets the inital index number to show.

    marker_radius
        Optional. Sets the marker radius as defined by matplotlib.

    axis_length
        Optional. Sets the rigid body size in meters.

    zoom
        Optional. Sets the initial camera zoom.

    azimuth
        Optional. Sets the initial camera azimuth in radians.

    elevation
        Optional. Sets the initial camera elevation in radians.

    translation
        Optional. Sets the initial camera translation (panning).

    target
        Optional. Sets the camera target in meters.

    track
        Optional. False to keep the scene static, True to track the last
        selected marker when changing index.

    perspective
        Optional. True to draw the scene using perspective, False to draw the
        scene orthogonally.

    """

    def __init__(
        self,
        *ts: TimeSeries,
        interconnections: Dict[str, Dict[str, Any]] = {},
        current_index: int = 0,
        marker_radius: float = 0.008,
        axis_length: float = 0.1,
        axis_width: float = 3.0,
        interconnection_width: float = 1.5,
        zoom: float = 1.0,
        azimuth: float = 0.0,
        elevation: float = 0.2,
        translation: Union[Sequence[float], np.ndarray] = (0.0, 0.0),
        target: Union[Sequence[float], np.ndarray] = (0.0, 0.0, 0.0),
        track: bool = False,
        perspective: bool = True,
        **kwargs,
    ):

        # Allow older variable names
        if "segments" in kwargs and interconnections == {}:
            interconnections = kwargs["segments"]
        if "segment_width" in kwargs:
            interconnection_width = kwargs["segment_width"]
        if "current_frame" in kwargs:
            current_index = kwargs["current_frame"]

        # ---------------------------------------------------------------
        # Set self.n_indexes and self.time, and verify that we have at least
        # markers or rigid bodies.
        self.time = ts[0].time
        self.n_indexes = len(ts[0].time)

        # ---------------------------------------------------------------
        # Assign the markers and rigid bodies
        self.markers = TimeSeries()
        self.rigid_bodies = TimeSeries()

        for one_ts in ts:
            for key in one_ts.data:
                if one_ts.data[key].shape[1:] == (4,):
                    self.markers.data[key] = one_ts.data[key]
                    if key in one_ts.data_info:
                        self.markers.data_info[key] = one_ts.data_info[key]

                elif one_ts.data[key].shape[1:] == (4, 4):

                    # Add the frames
                    self.rigid_bodies.data[key] = one_ts.data[key]
                    if key in one_ts.data_info:
                        self.rigid_bodies.data_info[key] = one_ts.data_info[
                            key
                        ]

                    # Add the frame origins as "markers"
                    key_label = f"[{key}] origin"
                    self.markers.data[key_label] = one_ts.data[key][:, :, 3]
                    if key in one_ts.data_info:
                        self.markers.data_info[key_label] = one_ts.data_info[
                            key
                        ]

                else:
                    raise ValueError(
                        "TimeSeries data must be of shape Nx4 "
                        "(markers) or Nx4x4 (rigid bodies)"
                    )
        self._select_none()
        self.last_selected_marker = ""

        # Add the origin to the rigid bodies
        self.rigid_bodies.data["Global"] = np.repeat(
            np.eye(4, 4)[np.newaxis, :, :], self.n_indexes, axis=0
        )

        # ---------------------------------------------------------------
        # Assign the interconnections
        self.interconnections = interconnections

        # ---------------------------------------------------------------
        # Other initalizations
        self.current_index = current_index
        self.marker_radius = marker_radius
        self.axis_length = axis_length
        self.axis_width = axis_width
        self.interconnection_width = interconnection_width
        self.zoom = zoom
        self.azimuth = azimuth
        self.elevation = elevation
        self.target = target
        self.track = track
        self.translation = translation
        self.perspective = perspective
        self.playback_speed = 1.0  # 0.0 to show all samples
        #  self.anim = None  # Will initialize in _create_figure

        self.objects = dict()  # type: Dict[str, Any]
        self._colors = ["r", "g", "b", "y", "c", "m", "w"]
        self.objects["PlotMarkers"] = dict()
        for color in self._colors:
            self.objects["PlotMarkers"][color] = None  # Not selected
            self.objects["PlotMarkers"][color + "s"] = None  # Selected
        self.objects["PlotRigidBodiesX"] = None
        self.objects["PlotRigidBodiesY"] = None
        self.objects["PlotRigidBodiesZ"] = None
        self.objects["PlotGroundPlane"] = None
        self.objects["PlotInterconnections"] = dict()

        self.objects["Figure"] = None
        self.objects["Axes"] = None
        self.objects["Help"] = None

        self.state = dict()  # type: Dict[str, Any]
        self.state["ShiftPressed"] = False
        self.state["MouseLeftPressed"] = False
        self.state["MouseMiddlePressed"] = False
        self.state["MouseRightPressed"] = False
        self.state["MousePositionOnPress"] = (0.0, 0.0)
        self.state["MousePositionOnMiddlePress"] = (0.0, 0.0)
        self.state["MousePositionOnRightPress"] = (0.0, 0.0)
        self.state["TranslationOnMousePress"] = (0.0, 0.0)
        self.state["AzimutOnMousePress"] = 0.0
        self.state["ElevationOnMousePress"] = 0.0
        self.state["SelfTimeOnPlay"] = self.time[0]
        self.state["SystemTimeOnLastUpdate"] = time.time()

        self._help_text = """
            ktk.Player help
            ----------------------------------------------------
            KEYBOARD COMMANDS
            show/hide this help : h
            previous index      : left
            next index          : right
            previous second     : shift+left
            next second         : shift+right
            play/pause          : space
            2x playback speed   : +
            0.5x playback speed : -
            toggle track        : t
            toggle perspective  : d (depth)
            ----------------------------------------------------
            MOUSE COMMANDS
            select a marker     : left-click
            3d rotate           : left-drag
            pan                 : middle-drag or shift+left-drag
            zoom                : right-drag or wheel
            """

        self._create_figure()
        self._create_interconnections()
        self._create_markers()
        self._create_ground_plane()
        self._first_refresh()

    def __dir__(self):
        return directory(Player.__dict__)

    def _create_figure(self) -> None:
        """Create the player's figure."""
        # Create the figure and axes
        self.objects["Figure"], self.objects["Axes"] = plt.subplots(
            num=None, figsize=(12, 9), facecolor="k", edgecolor="w"
        )

        # Remove the toolbar
        try:  # Try, setVisible method not always there
            self.objects["Figure"].canvas.toolbar.setVisible(False)
        except AttributeError:
            pass

        plt.tight_layout()

        # Add the title
        title_obj = plt.title("Player")
        plt.setp(title_obj, color=[0, 1, 0])  # Set a green title

        # Remove the background for faster plotting
        self.objects["Axes"].set_axis_off()

        # Add the animation timer
        self.anim = animation.FuncAnimation(
            self.objects["Figure"], self._on_timer, interval=33
        )  # 30 ips
        self.running = False

        # Connect the callback functions
        self.objects["Figure"].canvas.mpl_connect("pick_event", self._on_pick)
        self.objects["Figure"].canvas.mpl_connect(
            "key_press_event", self._on_key
        )
        self.objects["Figure"].canvas.mpl_connect(
            "key_release_event", self._on_release
        )
        self.objects["Figure"].canvas.mpl_connect(
            "scroll_event", self._on_scroll
        )
        self.objects["Figure"].canvas.mpl_connect(
            "button_press_event", self._on_mouse_press
        )
        self.objects["Figure"].canvas.mpl_connect(
            "button_release_event", self._on_mouse_release
        )
        self.objects["Figure"].canvas.mpl_connect(
            "motion_notify_event", self._on_mouse_motion
        )

    def _create_interconnections(self) -> None:
        """Create the interconnections plots in the player's figure."""
        if self.interconnections is not None:
            for interconnection in self.interconnections:
                self.objects["PlotInterconnections"][
                    interconnection
                ] = self.objects["Axes"].plot(
                    np.nan,
                    np.nan,
                    "-",
                    c=self.interconnections[interconnection]["Color"],
                    linewidth=self.interconnection_width,
                )[
                    0
                ]

    def _create_markers(self) -> None:
        """Create the markers plots in the player's figure."""
        colors = {
            "r": [1, 0, 0],
            "g": [0, 1, 0],
            "b": [0.3, 0.3, 1],
            "y": [1, 1, 0],
            "m": [1, 0, 1],
            "c": [0, 1, 1],
            "w": [0.8, 0.8, 0.8],
        }

        for color in self._colors:
            self.objects["PlotMarkers"][color] = self.objects["Axes"].plot(
                np.nan,
                np.nan,
                ".",
                c=colors[color],
                markersize=4,
                pickradius=5,
                picker=True,
            )[0]
        for color in self._colors:
            self.objects["PlotMarkers"][color + "s"] = self.objects[
                "Axes"
            ].plot(np.nan, np.nan, ".", c=colors[color], markersize=12)[0]

    def _create_ground_plane(self) -> None:
        # Create the ground plane matrix
        gp_size = 30  # blocks
        gp_div = 4  # blocks per meter
        gp_x = np.block(
            [
                np.tile(
                    [-gp_size / gp_div, gp_size / gp_div, np.nan], gp_size
                ),
                np.repeat(
                    np.linspace(-gp_size / gp_div, gp_size / gp_div, gp_size),
                    3,
                ),
            ]
        )
        gp_y = np.zeros(6 * gp_size)
        gp_z = np.block(
            [
                np.repeat(
                    np.linspace(-gp_size / gp_div, gp_size / gp_div, gp_size),
                    3,
                ),
                np.tile(
                    [-gp_size / gp_div, gp_size / gp_div, np.nan], gp_size
                ),
            ]
        )
        gp_1 = np.ones(6 * gp_size)
        self._ground_plane = np.hstack(
            [
                gp_x[:, np.newaxis],
                gp_y[:, np.newaxis],
                gp_z[:, np.newaxis],
                gp_1[:, np.newaxis],
            ]
        )

    def _first_refresh(self) -> None:
        """Draw the stuff and set the axis size."""
        self._update_plots()
        plt.axis([-1.5, 1.5, -1, 1])

    def _get_projection(self, points_3d: np.ndarray) -> np.ndarray:
        """
        Get a 3d --> 2d projection of a list of points.

        The method uses the class's camera variables to project a list of
        3d points onto a 2d canvas.

        Parameters
        ----------
        points_3d
            Nx4 array, where the first dimension is the number of points
            and  the second dimension is (x, y, z, 1).

        Returns
        -------
            Nx2 array, where the first dimension is the number of points and
            the second dimension is (x, y) to be ploted on a 2d graphic.

        """
        # ------------------------------------------------------------
        # Create the rotation matrix to convert the lab's coordinates
        # (x anterior, y up, z right) to the camera coordinates (x right,
        # y up, z deep)

        R = (
            np.array(
                [
                    [2 * self.zoom, 0, 0, 0],
                    [0, 2 * self.zoom, 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1],
                ]
            )
            @ np.array(
                [
                    [1, 0, 0, self.translation[0]],  # Pan
                    [0, 1, 0, self.translation[1]],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1],
                ]
            )
            @ np.array(
                [
                    [1, 0, 0, 0],
                    [0, cos(-self.elevation), sin(self.elevation), 0],
                    [0, sin(-self.elevation), cos(-self.elevation), 0],
                    [0, 0, 0, 1],
                ]
            )
            @ np.array(
                [
                    [cos(-self.azimuth), 0, sin(self.azimuth), 0],
                    [0, 1, 0, 0],
                    [sin(-self.azimuth), 0, cos(-self.azimuth), 0],
                    [0, 0, 0, 1],
                ]
            )
            @ np.array(
                [
                    [1, 0, 0, -self.target[0]],  # Rotate around target
                    [0, 1, 0, -self.target[1]],
                    [0, 0, -1, self.target[2]],
                    [0, 0, 0, 1],
                ]
            )
        )

        # Add a first dimension to R and match first dimension of points_3d
        R = np.repeat(R[np.newaxis, :], points_3d.shape[0], axis=0)

        # Rotate points.
        rotated_points_3d = R @ points_3d[:, :, np.newaxis]
        rotated_points_3d = rotated_points_3d[:, :, 0]

        # Apply perspective.
        if self.perspective is True:
            # This uses an ugly magical constant but it works fine for now.
            denom = rotated_points_3d[:, 2] / 10 + 5
            rotated_points_3d[:, 0] = rotated_points_3d[:, 0] / denom
            rotated_points_3d[:, 1] = rotated_points_3d[:, 1] / denom
            with np.errstate(invalid="ignore"):
                to_remove = denom < 1e-12
            rotated_points_3d[to_remove, 0] = np.nan
            rotated_points_3d[to_remove, 1] = np.nan
        else:
            # Scale to match the point of view
            rotated_points_3d /= 5

        # Return only x and y
        return rotated_points_3d[:, 0:2]

    def _update_markers_and_interconnections(self) -> None:
        # Get a Nx4 matrices of every marker at the current index
        markers = self.markers
        if markers is None:
            return
        else:
            n_markers = len(markers.data)

        markers_data = dict()  # Used to draw the markers with different colors
        interconnection_markers = dict()  # Used to draw the interconnections

        for color in self._colors:
            markers_data[color] = np.empty([n_markers, 4])
            markers_data[color][:] = np.nan

            markers_data[color + "s"] = np.empty([n_markers, 4])
            markers_data[color + "s"][:] = np.nan

        if n_markers > 0:
            for i_marker, marker in enumerate(markers.data):

                # Get this marker's color
                if (
                    marker in markers.data_info
                    and "Color" in markers.data_info[marker]
                ):
                    color = markers.data_info[marker]["Color"]
                else:
                    color = "w"

                these_coordinates = markers.data[marker][self.current_index]
                markers_data[color][i_marker] = these_coordinates
                interconnection_markers[marker] = these_coordinates

        # Update the markers plot
        for color in self._colors:
            # Unselected markers
            markers_data[color] = self._get_projection(markers_data[color])
            self.objects["PlotMarkers"][color].set_data(
                markers_data[color][:, 0], markers_data[color][:, 1]
            )

            # Selected markers
            markers_data[color + "s"] = self._get_projection(
                markers_data[color + "s"]
            )
            self.objects["PlotMarkers"][color + "s"].set_data(
                markers_data[color + "s"][:, 0],
                markers_data[color + "s"][:, 1],
            )

        # Draw the interconnections
        if self.interconnections is not None:
            for interconnection in self.interconnections:
                n_links = len(self.interconnections[interconnection]["Links"])
                coordinates = np.empty((3 * n_links, 4))
                coordinates[:] = np.nan
                for i_link in range(n_links):
                    marker1 = self.interconnections[interconnection]["Links"][
                        i_link
                    ][0]
                    marker2 = self.interconnections[interconnection]["Links"][
                        i_link
                    ][1]
                    if marker1 in interconnection_markers:
                        coordinates[3 * i_link] = interconnection_markers[
                            marker1
                        ]
                    else:
                        coordinates[3 * i_link] = np.nan

                    if marker2 in interconnection_markers:
                        coordinates[3 * i_link + 1] = interconnection_markers[
                            marker2
                        ]
                    else:
                        coordinates[3 * i_link + 1] = np.nan

                coordinates = self._get_projection(coordinates)

                self.objects["PlotInterconnections"][interconnection].set_data(
                    coordinates[:, 0], coordinates[:, 1]
                )

    def _update_plots(self) -> None:
        """Update the plots, or draw it if not plot has been drawn before."""
        self._update_markers_and_interconnections()

        # Get three (3N)x4 matrices (for x, y and z lines) for the rigid bodies
        # at the current index
        rigid_bodies = self.rigid_bodies
        n_rigid_bodies = len(rigid_bodies.data)
        rbx_data = np.empty([n_rigid_bodies * 3, 4])
        rby_data = np.empty([n_rigid_bodies * 3, 4])
        rbz_data = np.empty([n_rigid_bodies * 3, 4])

        for i_rigid_body, rigid_body in enumerate(rigid_bodies.data):
            # Origin
            rbx_data[i_rigid_body * 3] = rigid_bodies.data[rigid_body][
                self.current_index, :, 3
            ]
            rby_data[i_rigid_body * 3] = rigid_bodies.data[rigid_body][
                self.current_index, :, 3
            ]
            rbz_data[i_rigid_body * 3] = rigid_bodies.data[rigid_body][
                self.current_index, :, 3
            ]
            # Direction
            rbx_data[i_rigid_body * 3 + 1] = rigid_bodies.data[rigid_body][
                self.current_index
            ] @ np.array([self.axis_length, 0, 0, 1])
            rby_data[i_rigid_body * 3 + 1] = rigid_bodies.data[rigid_body][
                self.current_index
            ] @ np.array([0, self.axis_length, 0, 1])
            rbz_data[i_rigid_body * 3 + 1] = rigid_bodies.data[rigid_body][
                self.current_index
            ] @ np.array([0, 0, self.axis_length, 1])
            # NaN to cut the line between the different rigid bodies
            rbx_data[i_rigid_body * 3 + 2] = np.repeat(np.nan, 4)
            rby_data[i_rigid_body * 3 + 2] = np.repeat(np.nan, 4)
            rbz_data[i_rigid_body * 3 + 2] = np.repeat(np.nan, 4)

        # Update the ground plane
        gp = self._get_projection(self._ground_plane)
        if self.objects["PlotGroundPlane"] is None:  # Create the plot
            self.objects["PlotGroundPlane"] = self.objects["Axes"].plot(
                gp[:, 0], gp[:, 1], c=[0.3, 0.3, 0.3], linewidth=1
            )[0]
        else:  # Update the plot
            self.objects["PlotGroundPlane"].set_data(gp[:, 0], gp[:, 1])

        # Create or update the rigid bodies plot
        rbx_data = self._get_projection(rbx_data)
        rby_data = self._get_projection(rby_data)
        rbz_data = self._get_projection(rbz_data)
        if self.objects["PlotRigidBodiesX"] is None:  # Create the plot
            self.objects["PlotRigidBodiesX"] = self.objects["Axes"].plot(
                rbx_data[:, 0],
                rbx_data[:, 1],
                c="r",
                linewidth=self.axis_width,
            )[0]
            self.objects["PlotRigidBodiesY"] = self.objects["Axes"].plot(
                rby_data[:, 0],
                rby_data[:, 1],
                c="g",
                linewidth=self.axis_width,
            )[0]
            self.objects["PlotRigidBodiesZ"] = self.objects["Axes"].plot(
                rbz_data[:, 0],
                rbz_data[:, 1],
                c="b",
                linewidth=self.axis_width,
            )[0]
        else:  # Update the plot
            self.objects["PlotRigidBodiesX"].set_data(
                rbx_data[:, 0], rbx_data[:, 1]
            )
            self.objects["PlotRigidBodiesY"].set_data(
                rby_data[:, 0], rby_data[:, 1]
            )
            self.objects["PlotRigidBodiesZ"].set_data(
                rbz_data[:, 0], rbz_data[:, 1]
            )

        # Update the window title
        try:
            self.objects["Figure"].canvas.manager.set_window_title(
                f"{self.current_index}/{self.n_indexes}: "
                + "%2.2f s." % self.time[self.current_index]
            )
        except AttributeError:
            pass

        self.objects["Figure"].canvas.draw()

    def _set_new_target(
        self, target: Union[Sequence[float], np.ndarray]
    ) -> None:
        """Set new target and adapts translation and zoom consequently."""
        if np.sum(np.isnan(target)) > 0:
            return
        initial_translation = copy.deepcopy(self.translation)
        initial_zoom = copy.deepcopy(self.zoom)
        initial_target = copy.deepcopy(self.target)

        n_markers = len(self.markers.data)
        markers = np.empty((n_markers, 4))
        for i_marker, marker in enumerate(self.markers.data):
            markers[i_marker] = self.markers.data[marker][self.current_index]

        initial_projected_markers = self._get_projection(markers)
        # Do not consider markers that are not in the screen
        initial_projected_markers[
            initial_projected_markers[:, 0] < -1.5
        ] = np.nan
        initial_projected_markers[
            initial_projected_markers[:, 0] > 1.5
        ] = np.nan
        initial_projected_markers[
            initial_projected_markers[:, 1] < -1.0
        ] = np.nan
        initial_projected_markers[
            initial_projected_markers[:, 1] > 1.0
        ] = np.nan
        self.target = target

        def error_function(input):
            self.translation = input[0:2]
            self.zoom = input[2]
            new_projected_markers = self._get_projection(markers)
            error = np.nanmean(
                (initial_projected_markers - new_projected_markers) ** 2
            )
            return error

        res = optim.minimize(
            error_function, np.hstack((self.translation, self.zoom))
        )
        if res.success is False:
            self.translation = initial_translation
            self.zoom = initial_zoom
            self.target = initial_target

    # ------------------------------------
    # Helper functions
    def _set_index(self, index: int) -> None:
        """Set current index to a given index and update plots."""
        if index >= self.n_indexes:
            self.current_index = self.n_indexes - 1
        elif index < 0:
            self.current_index = 0
        else:
            self.current_index = index

        if self.track is True and self.markers is not None:
            new_target = self.markers.data[self.last_selected_marker][
                self.current_index
            ]
            if not np.isnan(np.sum(new_target)):
                self.target = new_target

    def _set_time(self, time: float) -> None:
        """Set current index to a given time and update plots."""
        index = int(np.argmin(np.abs(self.time - time)))
        self._set_index(index)

    def _select_none(self) -> None:
        """Deselect every markers."""
        if self.markers is not None:
            for marker in self.markers.data:
                try:
                    # Keep 1st character, remove the possible 's'
                    self.markers.data_info[marker][
                        "Color"
                    ] = self.markers.data_info[marker]["Color"][0]
                except KeyError:
                    self.markers = self.markers.add_data_info(
                        marker, "Color", "w"
                    )

    def close(self) -> None:
        """Close the Player and its associated window."""
        plt.close(self.objects["Figure"])
        self.objects = {}

    # ------------------------------------
    # Callbacks
    def _on_close(self, _) -> None:  # pragma: no cover
        # Release all references to objects
        self.close()

    def _on_timer(self, _) -> None:  # pragma: no cover
        if self.running is True:
            # We check self.running because the garbage collector may take time
            # before deleting the animation timer, and unreferencing the
            # animation timer is the recommended way to deactivate a timer.
            """Callback for the animation timer object."""
            current_index = self.current_index
            self._set_time(
                self.time[self.current_index]
                + self.playback_speed
                * (time.time() - self.state["SystemTimeOnLastUpdate"])
            )
            if current_index == self.current_index:
                # The time wasn't enough to advance a frame. Articifically
                # advance a frame.
                self._set_index(self.current_index + 1)
            self.state["SystemTimeOnLastUpdate"] = time.time()

            self._update_plots()
        else:
            self.anim.event_source.stop()

    def _on_pick(self, event):  # pragma: no cover
        """Callback for marker selection."""
        if event.mouseevent.button == 1:
            index = event.ind
            selected_marker = list(self.markers.data.keys())[index[0]]
            self.objects["Axes"].set_title(selected_marker)

            # Mark selected
            self._select_none()
            self.markers.data_info[selected_marker]["Color"] = (
                self.markers.data_info[selected_marker]["Color"][0] + "s"
            )

            # Set as new target
            self.last_selected_marker = selected_marker
            self._set_new_target(
                self.markers.data[selected_marker][self.current_index]
            )
            marker_position = self.markers.data[selected_marker][
                self.current_index
            ]

            self._update_plots()

    def _on_key(self, event):  # pragma: no cover
        """Callback for keyboard key pressed."""
        if event.key == " ":
            if self.running is False:
                self.state["SystemTimeOnLastUpdate"] = time.time()
                self.state["SelfTimeOnPlay"] = self.time[self.current_index]
                self.running = True
                self.anim.event_source.start()
            else:
                self.running = False
                self.anim.event_source.stop()

        elif event.key == "left":
            self._set_index(self.current_index - 1)

        elif event.key == "shift+left":
            self._set_time(self.time[self.current_index] - 1)

        elif event.key == "right":
            self._set_index(self.current_index + 1)

        elif event.key == "shift+right":
            self._set_time(self.time[self.current_index] + 1)

        elif event.key == "-":
            self.playback_speed /= 2
            self.objects["Axes"].set_title(
                f"Playback set to {self.playback_speed}x"
            )

        elif event.key == "+":
            self.playback_speed *= 2
            self.objects["Axes"].set_title(
                f"Playback set to {self.playback_speed}x"
            )

        elif event.key == "h":
            if self.objects["Help"] is None:
                self.objects["Help"] = self.objects["Axes"].text(
                    -1.5,
                    -1,
                    self._help_text,
                    color=[0, 1, 0],
                    fontfamily="monospace",
                )
            else:
                self.objects["Help"].remove()
                self.objects["Help"] = None

        elif event.key == "d":
            self.perspective = not self.perspective
            if self.perspective is True:
                self.objects["Axes"].set_title(f"Camera set to perspective")
            else:
                self.objects["Axes"].set_title(f"Camera set to orthogonal")

        elif event.key == "t":
            self.track = not self.track
            if self.track is True:
                self.objects["Axes"].set_title(f"Marker tracking activated")
            else:
                self.objects["Axes"].set_title(f"Marker tracking deactivated")

        elif event.key == "shift":
            self.state["ShiftPressed"] = True

        self._update_plots()

    def _on_release(self, event):  # pragma: no cover
        if event.key == "shift":
            self.state["ShiftPressed"] = False

    def _on_scroll(self, event):  # pragma: no cover
        if event.button == "up":
            self.zoom *= 1.05
        elif event.button == "down":
            self.zoom /= 1.05
        self._update_plots()

    def _on_mouse_press(self, event):  # pragma: no cover

        if len(self.last_selected_marker) > 0:
            self._set_new_target(
                self.markers.data[self.last_selected_marker][
                    self.current_index
                ]
            )

        self.state["TranslationOnMousePress"] = self.translation
        self.state["AzimutOnMousePress"] = self.azimuth
        self.state["ElevationOnMousePress"] = self.elevation
        self.state["ZoomOnMousePress"] = self.zoom
        self.state["MousePositionOnPress"] = (event.x, event.y)
        if event.button == 1:
            self.state["MouseLeftPressed"] = True
        elif event.button == 2:
            self.state["MouseMiddlePressed"] = True
        elif event.button == 3:
            self.state["MouseRightPressed"] = True

    def _on_mouse_release(self, event):  # pragma: no cover
        if event.button == 1:
            self.state["MouseLeftPressed"] = False
        elif event.button == 2:
            self.state["MouseMiddlePressed"] = False
        elif event.button == 3:
            self.state["MouseRightPressed"] = False

    def _on_mouse_motion(self, event):  # pragma: no cover
        # Pan:
        if (
            self.state["MouseLeftPressed"] and self.state["ShiftPressed"]
        ) or self.state["MouseMiddlePressed"]:
            self.translation = (
                self.state["TranslationOnMousePress"][0]
                + (event.x - self.state["MousePositionOnPress"][0])
                / (100 * self.zoom),
                self.state["TranslationOnMousePress"][1]
                + (event.y - self.state["MousePositionOnPress"][1])
                / (100 * self.zoom),
            )
            self._update_plots()

        # Rotation:
        elif self.state["MouseLeftPressed"] and not self.state["ShiftPressed"]:
            self.azimuth = (
                self.state["AzimutOnMousePress"]
                - (event.x - self.state["MousePositionOnPress"][0]) / 250
            )
            self.elevation = (
                self.state["ElevationOnMousePress"]
                - (event.y - self.state["MousePositionOnPress"][1]) / 250
            )
            self._update_plots()

        # Zoom:
        elif self.state["MouseRightPressed"]:
            self.zoom = (
                self.state["ZoomOnMousePress"]
                + (event.y - self.state["MousePositionOnPress"][1]) / 250
            )
            self._update_plots()

    def to_html5(
        self,
        start_index: Optional[int] = None,
        stop_index: Optional[int] = None,
        start_time: Optional[int] = None,
        stop_time: Optional[int] = None,
        **kwargs,
    ):
        """
        Create an html5 video for displaying in Jupyter notebooks.

        Parameters
        ----------
        start_index
            Start index in the exported video. Default is the beginning.

        stop_index
            Stop index in the exported video. Negative numbers to count from
            the end. Default is the end.

        start_time
            Start time in the exported video. Default is the beginning.
            Ignored if start_index is set.

        stop_time
            Stop time in the exported video. Default is the end.
            Ignored if stop_index is set.

        Returns
        -------
        A matplotlib animation object for displaying as html5.

        Warnings
        --------
        This function, which has been introduced in 0.4, is still experimental
        and may change signature or behaviour in the future.

        """
        # Allow old arguments
        if "start_frame" in kwargs:
            start_index = kwargs["start_frame"]
        if "stop_frame" in kwargs:
            stop_index = kwargs["stop_frame"]

        mpl.rcParams["animation.html"] = "html5"

        if start_index is None:
            if start_time is None:
                start_index = 0
            else:
                start_index = int(np.argmin(np.abs(self.time - start_time)))

        if stop_index is None:
            if stop_time is None:
                stop_index = -1
            else:
                stop_index = int(np.argmin(np.abs(self.time - stop_time)))

        if stop_index < 0:
            stop_index = self.n_indexes + stop_index + 1

        self.playback_speed = 0
        self.objects["Figure"].set_size_inches(6, 4.5)  # Half size
        self._set_index(start_index)
        self.running = True
        self.anim.save_count = stop_index - start_index
        self.anim.event_source.start()
        plt.close(self.objects["Figure"])
        return self.anim



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
kineticstoolkit._repr.py
------------------------

Format the console output of dictionaries and classes with attributes.

This module formats the console output of dict in ipython, so that instead
of just using repr(), it displays a nicer list of keys with abbreviated
values if required, so that there is a maximum of one key per line. This is
very useful for nested dicts, since their repr() representation is recursive
and becomes unmanagable when the dict becomes larger.

It also provides helper functions to nicely format the repr() of data classes.

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


import numpy as np
from typing import Dict, Any


def _format_dict_entries(the_dict: Any, quotes: bool = True) -> str:
    """
    Format a dict nicely on screen.

    This function makes every element of a dict appear on a separate line,
    with each keys right aligned:
        {
           'key1': value1
           'key2': value2
        'longkey': value3
        }

    Parameters
    ----------
    the_dict:
        The input dictionary
    quotes:
        False to remove quotes from keys when they are strings. Default is
        True.

    Returns
    -------
    A string that should be shown by the __repr__ method.

    """
    max_width = 79  # How many characters should we print
    out = ""

    widest = 0
    # Find the widest key name
    for key in the_dict:
        key_label = repr(key)
        if quotes is False and isinstance(key_label, str):
            key_label = key_label[1:-1]
        widest = max(widest, len(key_label))

    # Print each key value
    for key in the_dict:

        key_label = repr(key)
        if quotes is False and isinstance(key_label, str):
            key_label = key_label[1:-1]
        key_label = " " * (widest - len(key_label)) + key_label

        value = the_dict[key]

        # Print the value
        if isinstance(value, dict):
            value_label = "<dict with " + str(len(value)) + " entries>"
        elif isinstance(value, list):
            value_label = "<list of " + str(len(value)) + " items>"
        elif isinstance(value, np.ndarray):
            value_label = "<array of shape " + str(np.shape(value)) + ">"
        else:
            value_label = repr(value)

        # Remove line breaks and multiple-spaces
        value_label = " ".join(value_label.split())

        # Printout
        to_print = f"    {key_label}: {value_label}"

        if len(to_print) > max_width:
            to_print = to_print[0 : max_width - 3] + "..."

        out += to_print
        out += "\n"

    return out


def _format_class_attributes(obj):
    """
    Format a class that has attributes nicely on screen.

    This class lists every attribute of a class on a separate line, using the
    _format_dict_entries function:

        ClassName with attributes:
           'attribute1' : value1
           'attribute2' : value2
        'longattribute' : value3

    Parameters
    ----------
    obj: Any
        The class instance.

    Returns
    -------
    A string that should be shown by the __repr__ method.

    """
    # Return the type of class (header)
    class_name = type(obj).__name__
    out = class_name + " with attributes:\n"

    # Return the list of attributes
    out += _format_dict_entries(obj.__dict__, quotes=False)
    return out


def _ktk_format_dict(value, p, cycle):
    """Format a dict nicely on screen in ipython."""
    try:
        get_ipython()

        if cycle:
            p.pretty("...")
        else:
            p.text("{\n")
            p.text(_format_dict_entries(value))
            p.text("}")

    except:
        p.text(repr(value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2020 Félix Chénier

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provide decorators for Kinetics Toolkit's functions.

The following decorator can be used on each Kinetics Toolkit's's function:

    - @unstable:
        Documented only in the development version.

    - @deprecated(since: str, until: str, details: str):
        Documented but deprecated function in a release.

    - @dead:
        Undocumented, deprecated function in the development version.

    Each of these decorators add the _include_in_dir property to the decorated
    function. The provided function ``directory`` looks at these properties
    to return a custom __dir__ to Kinematics Toolkit's classes. See such class
    for example.

"""

__author__ = "Félix Chénier"
__copyright__ = "Copyright (C) 2020 Félix Chénier"
__email__ = "chenier.felix@uqam.ca"
__license__ = "Apache 2.0"


from functools import wraps
import warnings
import textwrap
import kineticstoolkit.config
from typing import Dict, List, Any


def _inject_in_docstring(docstring: str, text: str) -> str:
    """Inject a string into the top of a docstring, after line 1."""
    if docstring == "" or docstring is None:
        return text
    result = []
    splitted = textwrap.dedent(docstring).split("\n")
    first_line_done = False
    for line in splitted:
        if not first_line_done:
            if len(line) != 0:
                first_line_done = True
                result.append(line)
                for text_line in text.split("\n"):
                    result.append(text_line)
            else:
                result.append(line)
        else:
            result.append(line)
    return "\n".join(result)


def deprecated(since: str, until: str, details: str):
    """
    Decorate deprecated Kinetics Toolkit's functions.

    Generates a FutureWarning and adds a warning section to its docstring.
    These functions are included in API documentation.

    """

    def real_decorator(func):
        func_name = func.__name__
        string = (
            f"The function {func_name} is deprecated since "
            f"{since} and is scheduled to be removed "
            f"in {until}. {details}"
        )

        # Ensure the decorated function keeps its metadata
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(string, category=FutureWarning, stacklevel=2)
            # Call the function being decorated and return the result
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=FutureWarning)
                return func(*args, **kwargs)

        wrapper.__doc__ = _inject_in_docstring(
            func.__doc__, f"\nWarning\n-------\n{string}"
        )
        wrapper._is_deprecated = True
        return wrapper

    return real_decorator


def unstable(func):
    """
    Decorate unstable Kinetics Toolkit's functions.

    if kineticstoolkit.config.version == 'master':
        Adds this function to the main documentation.
    else:
        Generate a KTKUnstableWarning on use.

    Also adds a warning section to its docstring.

    """
    func_name = func.__name__
    string = (
        f"The function {func_name} is unstable, which means it may not "
        f"be tested or settled yet. Please avoid using this function "
        f"in production code."
    )

    # Ensure the decorated function keeps its metadata
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Call the function being decorated and return the result
        return func(*args, **kwargs)

    wrapper._is_unstable = True

    wrapper.__doc__ = _inject_in_docstring(
        func.__doc__, f"\nWarning\n-------\n{string}"
    )
    return wrapper


def dead(since: str, until: str, details=""):
    """
    Decorate dead Kinetics Toolkit's functions.

    Does not add this function to the main documentation and generates a
    FutureWarning on use. Also adds a warning section to its docstring.

    Parameter listing is a list of attributes of the module that will be
    returned by the module's or class' __dir__ function.
    """

    def real_decorator(func):
        func_name = func.__name__
        string = (
            f"The function {func_name} is deprecated since "
            f"{since} and is scheduled to be removed "
            f"in {until}. {details}"
        )

        # Ensure the decorated function keeps its metadata
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(string, category=FutureWarning, stacklevel=2)
            # Call the function being decorated and return the result
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=FutureWarning)
                return func(*args, **kwargs)

        wrapper.__doc__ = _inject_in_docstring(
            func.__doc__, f"\nWarning\n-------\n{string}"
        )
        wrapper._is_dead = True
        return wrapper

    return real_decorator


def directory(module_locals: Dict[str, Any]) -> List[str]:
    """
    Return the module's public directory for dir function.

    Parameters
    ----------
    module_locals
        The module's locals as generated by the locals() function.

    Returns
    -------
    List of public objects.

    """
    dir_ = []
    for key in module_locals:

        if key.startswith("_"):
            continue

        try:
            if (
                "_is_unstable" in module_locals[key].__dict__
                and module_locals[key].__dict__["_is_unstable"] is True
                and kineticstoolkit.config.version != "master"
            ):
                continue

            if (
                "_is_dead" in module_locals[key].__dict__
                and module_locals[key].__dict__["_is_dead"] is True
            ):
                continue

        except AttributeError:
            continue

        dir_.append(key)

    return dir_


def __dir__():
    return []
